from dataclasses import dataclass
from typing import List, Optional

import numpy as np
import torch
from dreifus.image import Img
from matplotlib.cm import get_cmap
from torch.nn import MSELoss

from visage.evaluator.lpips_evaluator import LPIPSEvaluator
from visage.evaluator.psnr_evaluator import PSNREvaluator
from visage.evaluator.ssim_evaluator import SSIMEvaluator, MultiScaleSSIMEvaluator


@dataclass
class PairedImageMetrics:
    psnr: float
    ssim: float
    lpips: float
    multi_scale_ssim: Optional[float] = None
    mse: Optional[float] = None

    def __add__(self, other: 'PairedImageMetrics') -> 'PairedImageMetrics':
        return PairedImageMetrics(
            psnr=self.psnr + other.psnr,
            ssim=self.ssim + other.ssim,
            lpips=self.lpips + other.lpips,
            multi_scale_ssim=self.multi_scale_ssim + other.multi_scale_ssim if self.multi_scale_ssim is not None and other.multi_scale_ssim is not None else None,
            mse=self.mse + other.mse if self.mse is not None and other.mse is not None else None,
        )

    def __sub__(self, other: 'PairedImageMetrics') -> 'PairedImageMetrics':
        return PairedImageMetrics(
            psnr=self.psnr - other.psnr,
            ssim=self.ssim - other.ssim,
            lpips=self.lpips - other.lpips,
            multi_scale_ssim=self.multi_scale_ssim - other.multi_scale_ssim if self.multi_scale_ssim is not None and other.multi_scale_ssim is not None else None,
            mse=self.mse - other.mse if self.mse is not None and other.mse is not None else None,
        )

    def __pow__(self, power: int, modulo=None)-> 'PairedImageMetrics':
        return PairedImageMetrics(
            psnr=self.psnr.__pow__(power, modulo),
            ssim=self.ssim.__pow__(power, modulo),
            lpips=self.lpips.__pow__(power, modulo),
            multi_scale_ssim=self.multi_scale_ssim.__pow__(power, modulo) if self.multi_scale_ssim is not None else None,
            mse=self.mse.__pow__(power, modulo) if self.mse is not None else None,
        )

    def __radd__(self, other) -> 'PairedImageMetrics':
        if other == 0:
            return self

        return self + other

    def __truediv__(self, scalar: float) -> 'PairedImageMetrics':
        return PairedImageMetrics(
            psnr=self.psnr / scalar,
            ssim=self.ssim / scalar,
            lpips=self.lpips / scalar,
            multi_scale_ssim=self.multi_scale_ssim / scalar if self.multi_scale_ssim is not None else None,
            mse=self.mse / scalar if self.mse is not None else None,
        )


class PairedImageEvaluator:

    def __init__(self, exclude_lpips: bool = False):
        self._psnr_evaluator = PSNREvaluator()
        self._ssim_evaluator = SSIMEvaluator()
        self._multi_scale_ssim_evaluator = MultiScaleSSIMEvaluator()
        if not exclude_lpips:
            self._lpips_evaluator = LPIPSEvaluator()
        self._mse_evaluator = MSELoss()

        self._exclude_lpips = exclude_lpips

    def evaluate(self, predictions: List[np.ndarray], targets: List[np.ndarray]) -> PairedImageMetrics:
        predictions_torch = [Img.from_numpy(prediction).to_torch().img.float() for prediction in predictions]
        targets_torch = [Img.from_numpy(target).to_torch().img.float() for target in targets]

        # Exclude targets that are all 0s
        proper_target_mask = [~((target == 0) | (target == 128)).all(axis=-1).all(axis=-1).all(axis=-1)
                              for target in targets]
        predictions_torch = [prediction for i, prediction in enumerate(predictions_torch) if proper_target_mask[i]]
        targets_torch = [target for i, target in enumerate(targets_torch) if proper_target_mask[i]]

        predictions_torch = torch.stack(predictions_torch)  # [B, 3, H, W]
        targets_torch = torch.stack(targets_torch)  # [B, 3, H, W]

        assert predictions_torch.min() >= 0
        assert predictions_torch.max() <= 1
        assert targets_torch.max() <= 1
        assert targets_torch.min() >= 0

        return self._evaluate(predictions_torch, targets_torch)

    def _evaluate(self, predictions_torch: torch.Tensor, targets_torch: torch.Tensor) -> PairedImageMetrics:
        psnr = self._psnr_evaluator(predictions_torch, targets_torch)
        ssim = self._ssim_evaluator(predictions_torch, targets_torch)
        multi_scale_ssim = self._multi_scale_ssim_evaluator(predictions_torch, targets_torch)
        if self._exclude_lpips:
            lpips = torch.tensor(-1)
        else:
            lpips = self._lpips_evaluator(predictions_torch, targets_torch)
        mse = self._mse_evaluator(predictions_torch, targets_torch)

        return PairedImageMetrics(psnr=psnr.item(), ssim=ssim.item(), lpips=lpips.item(), multi_scale_ssim=multi_scale_ssim.item(), mse=mse.item())

    def compute_error_maps(self, predictions: List[np.ndarray], targets: List[np.ndarray]) -> List[np.ndarray]:
        error_maps = []
        for prediction, target in zip(predictions, targets):
            error = np.linalg.norm(prediction / 255. - target / 255., axis=-1, ord=1)
            cmap = get_cmap('turbo')
            error_map = cmap(error)
            error_maps.append(error_map[..., :3])

        return error_maps



import torch
from torch import nn
from torchvision import transforms
from modnet.models.modnet import MODNet
import torch.nn.functional as F
class MODNetMatter:

    def __init__(self):
        ckpt_path = "C:/Users/kirschstein/Downloads/modnet_photographic_portrait_matting.ckpt"

        # define image to tensor transform
        self._transform = transforms.Compose(
            [
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
            ]
        )

        # create MODNet and load the pre-trained ckpt
        self._modnet = MODNet(backbone_pretrained=False)
        self._modnet = nn.DataParallel(self._modnet)

        if torch.cuda.is_available():
            self._modnet = self._modnet.cuda()
            weights = torch.load(ckpt_path)
        else:
            weights = torch.load(ckpt_path, map_location=torch.device('cpu'))
        self._modnet.load_state_dict(weights)
        self._modnet.eval()

    def parse(self, images: torch.Tensor) -> torch.Tensor:
        """

        :param images:
            Can be one or many images. Either in channel-first or channel-last format:
             - [B, 3, H, W]
             - [B, H, W, 3]
             - [3, H, W]
             - [H, W, 3]
            Images should be normalized to [0, 1]!
        :return:
            A matting mask of the same shape as the input, but the channel dimension collapsed.
            Example input: [B, 3, H, W] -> [B, H, W]
        """

        needs_squeeze = False
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
            needs_squeeze = True

        if images.shape[-1] == 3:
            # torchvision transforms requires images in Channel-first format
            images = images.permute(0, 3, 1, 2)

        # inference images
        images = self._transform(images)
        ref_size = 512

        # resize image for input
        im_b, im_c, im_h, im_w = images.shape
        if max(im_h, im_w) < ref_size or min(im_h, im_w) > ref_size:
            if im_w >= im_h:
                im_rh = ref_size
                im_rw = int(im_w / im_h * ref_size)
            elif im_w < im_h:
                im_rw = ref_size
                im_rh = int(im_h / im_w * ref_size)
        else:
            im_rh = im_h
            im_rw = im_w

        im_rw = im_rw - im_rw % 32
        im_rh = im_rh - im_rh % 32
        images = F.interpolate(images, size=(im_rh, im_rw), mode='area')

        # inference
        _, _, matte = self._modnet(images.cuda() if torch.cuda.is_available() else images, True)

        # resize and save matte
        matte = F.interpolate(matte, size=(im_h, im_w), mode='area')
        matte = matte.squeeze(-3)

        if needs_squeeze:
            matte = matte.squeeze(0)

        return matte


from typing import Dict

import torch
from torchvision import transforms
from torchvision.models.segmentation import deeplabv3_resnet101


class DeepLabV3FaceParser:

    def __init__(self):
        seg_net = deeplabv3_resnet101(pretrained=True, progress=False).to('cuda:0')
        seg_net.requires_grad_(False)
        seg_net.eval()

        self._seg_net = seg_net
        self._transform = transforms.Compose([
            transforms.Resize((512, 512)),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])

    def parse(self, images: torch.Tensor) -> torch.Tensor:
        """

        :param images:
            Can be one or many images. Either in channel-first or channel-last format:
             - [B, 3, H, W]
             - [B, H, W, 3]
             - [3, H, W]
             - [H, W, 3]
            Images should be normalized to [0, 1]!
        :return:
            A segmentation mask of the same shape as the input, but the channel dimension collapsed
            which indicates 1 out of 20 classes for each pixel.
            Example input: [B, 3, H, W] -> [B, H, W]
        """

        needs_squeeze = False
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
            needs_squeeze = True

        if images.shape[-1] == 3:
            # torchvision transforms requires images in Channel-first format
            images = images.permute(0, 3, 1, 2)

        images = self._transform(images)
        images = images.cuda()
        output = self._seg_net(images)['out']
        segmentation_masks = torch.nn.functional.softmax(output, dim=1)
        segmentation_masks = segmentation_masks.cpu().detach().argmax(1)

        if needs_squeeze:
            segmentation_masks = segmentation_masks.squeeze(0)

        return segmentation_masks

    @classmethod
    def get_label_mapping(cls) -> Dict[int, str]:

        return {0: 'background',
                1: 'aeroplane',
                2: 'bicycle',
                3: 'bird',
                4: 'boat',
                5: 'bottle',
                6: 'bus',
                7: 'car',
                8: 'cat',
                9: 'chair',
                10: 'cow',
                11: 'diningtable',
                12: 'dog',
                13: 'horse',
                14: 'motorbike',
                15: 'person',
                16: 'pottedplant',
                17: 'sheep',
                18: 'sofa',
                19: 'train',
                20: 'tvmonitor'}



# --------------------------------------------------------
# Faster R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------

import numpy as np
cimport numpy as np

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "gpu_nms.hpp":
    void _nms(np.int32_t*, int*, np.float32_t*, int, int, float, int)

def gpu_nms(np.ndarray[np.float32_t, ndim=2] dets, np.float thresh,
            np.int32_t device_id=0):
    cdef int boxes_num = dets.shape[0]
    cdef int boxes_dim = dets.shape[1]
    cdef int num_out
    cdef np.ndarray[np.int32_t, ndim=1] \
        keep = np.zeros(boxes_num, dtype=np.int32)
    cdef np.ndarray[np.float32_t, ndim=1] \
        scores = dets[:, 4]
    cdef np.ndarray[np.int_t, ndim=1] \
        order = scores.argsort()[::-1]
    cdef np.ndarray[np.float32_t, ndim=2] \
        sorted_dets = dets[order, :]
    _nms(&keep[0], &num_out, &sorted_dets[0, 0], boxes_num, boxes_dim, thresh, device_id)
    keep = keep[:num_out]
    return list(order[keep])



from dataclasses import dataclass
from typing import List

import numpy as np
import pyfvvdp


@dataclass
class PairedVideoMetric:
    jod: float

    def __add__(self, other: 'PairedVideoMetric') -> 'PairedVideoMetric':
        return PairedVideoMetric(
            jod=self.jod + other.jod,
        )

    def __radd__(self, other) -> 'PairedVideoMetric':
        if other == 0:
            return self

        return self + other

    def __truediv__(self, scalar: float) -> 'PairedVideoMetric':
        return PairedVideoMetric(
            jod=self.jod / scalar,
        )


class PairedVideoEvaluator:

    def __init__(self, fps: float):
        self._jod_evaluator = pyfvvdp.fvvdp(display_name='standard_4k', heatmap='threshold')
        self._evaluation_fps = fps

    def evaluate(self, prediction_video: List[np.ndarray], target_video: List[np.ndarray]) -> PairedVideoMetric:
        prediction_video = np.stack(prediction_video)  # [T, H, W, C]
        target_video = np.stack(target_video)  # [T, H, W, C]
        jod, _ = self._jod_evaluator.predict(prediction_video, target_video, dim_order="FHWC",
                                             frames_per_second=max(4.1, self._evaluation_fps))

        paired_video_metric = PairedVideoMetric(jod=jod.item())
        return paired_video_metric



# coding: utf-8

# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------

import os
import sys
from os.path import join as pjoin
import numpy as np
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext


def find_in_path(name, path):
    "Find a file in a search path"
    # adapted fom http://code.activestate.com/recipes/52224-find-a-file-given-a-search-path/
    for dir in path.split(os.pathsep):
        binpath = pjoin(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None


# Obtain the numpy include directory.  This logic works across numpy versions.
try:
    numpy_include = np.get_include()
except AttributeError:
    numpy_include = np.get_numpy_include()


# run the customize_compiler
class custom_build_ext(build_ext):
    def build_extensions(self):
        # customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)


ext_modules = [
    Extension(
        "visage.modules.FaceBoxesV2.utils.nms.cpu_nms",
        ["visage/modules/FaceBoxesV2/utils/nms/cpu_nms.pyx"],
        extra_compile_args=["-Wno-cpp", "-Wno-unused-function"] if sys.platform == 'linux' else None,
        include_dirs=[numpy_include]
    )
]

setup(
    name='mot_utils',
    ext_modules=ext_modules,
    # inject our custom trigger
    cmdclass={'build_ext': custom_build_ext},
)



from typing import List

import numpy as np

from visage.modules.FaceBoxesV2.faceboxes_detector import FaceBoxesDetector, DetectedBBox


class FaceBoxesV2(FaceBoxesDetector):

    def forward(self, image: np.ndarray, thresh: float = 0.6) -> List[DetectedBBox]:
        return self.detect(image, thresh=thresh)



import torch
from itertools import product as product
import numpy as np
from math import ceil


class PriorBox(object):
    def __init__(self, cfg, image_size=None, phase='train'):
        super(PriorBox, self).__init__()
        #self.aspect_ratios = cfg['aspect_ratios']
        self.min_sizes = cfg['min_sizes']
        self.steps = cfg['steps']
        self.clip = cfg['clip']
        self.image_size = image_size
        self.feature_maps = [[ceil(self.image_size[0]/step), ceil(self.image_size[1]/step)] for step in self.steps]

    def forward(self):
        anchors = []
        for k, f in enumerate(self.feature_maps):
            min_sizes = self.min_sizes[k]
            for i, j in product(range(f[0]), range(f[1])):
                for min_size in min_sizes:
                    s_kx = min_size / self.image_size[1]
                    s_ky = min_size / self.image_size[0]
                    if min_size == 32:
                        dense_cx = [x*self.steps[k]/self.image_size[1] for x in [j+0, j+0.25, j+0.5, j+0.75]]
                        dense_cy = [y*self.steps[k]/self.image_size[0] for y in [i+0, i+0.25, i+0.5, i+0.75]]
                        for cy, cx in product(dense_cy, dense_cx):
                            anchors += [cx, cy, s_kx, s_ky]
                    elif min_size == 64:
                        dense_cx = [x*self.steps[k]/self.image_size[1] for x in [j+0, j+0.5]]
                        dense_cy = [y*self.steps[k]/self.image_size[0] for y in [i+0, i+0.5]]
                        for cy, cx in product(dense_cy, dense_cx):
                            anchors += [cx, cy, s_kx, s_ky]
                    else:
                        cx = (j + 0.5) * self.steps[k] / self.image_size[1]
                        cy = (i + 0.5) * self.steps[k] / self.image_size[0]
                        anchors += [cx, cy, s_kx, s_ky]
        # back to torch land
        output = torch.Tensor(anchors).view(-1, 4)
        if self.clip:
            output.clamp_(max=1, min=0)
        return output



import os, cv2
import hdf5storage
import numpy as np
import sys

def process_300w(root_folder, folder_name, image_name, label_name, target_size):
    image_path = os.path.join(root_folder, folder_name, image_name)
    label_path = os.path.join(root_folder, folder_name, label_name)

    with open(label_path, 'r') as ff:
        anno = ff.readlines()[3:-1]
        anno = [x.strip().split() for x in anno]
        anno = [[int(float(x[0])), int(float(x[1]))] for x in anno]
        image = cv2.imread(image_path)
        image_height, image_width, _ = image.shape
        anno_x = [x[0] for x in anno]
        anno_y = [x[1] for x in anno]
        bbox_xmin = min(anno_x)
        bbox_ymin = min(anno_y)
        bbox_xmax = max(anno_x)
        bbox_ymax = max(anno_y)
        bbox_width = bbox_xmax - bbox_xmin
        bbox_height = bbox_ymax - bbox_ymin
        scale = 1.1 
        bbox_xmin -= int((scale-1)/2*bbox_width)
        bbox_ymin -= int((scale-1)/2*bbox_height)
        bbox_width *= scale
        bbox_height *= scale
        bbox_width = int(bbox_width)
        bbox_height = int(bbox_height)
        bbox_xmin = max(bbox_xmin, 0)
        bbox_ymin = max(bbox_ymin, 0)
        bbox_width = min(bbox_width, image_width-bbox_xmin-1)
        bbox_height = min(bbox_height, image_height-bbox_ymin-1)
        anno = [[(x-bbox_xmin)/bbox_width, (y-bbox_ymin)/bbox_height] for x,y in anno]

        bbox_xmax = bbox_xmin + bbox_width
        bbox_ymax = bbox_ymin + bbox_height
        image_crop = image[bbox_ymin:bbox_ymax, bbox_xmin:bbox_xmax, :]
        image_crop = cv2.resize(image_crop, (target_size, target_size))
        return image_crop, anno
        
def process_cofw(image, bbox, anno, target_size):
    image_height, image_width, _ = image.shape
    anno_x = anno[:29]
    anno_y = anno[29:58]
    ################################
    xmin, ymin, width, height = bbox
    xmax = xmin + width -1
    ymax = ymin + height -1
    ################################
    xmin = max(xmin, 0)
    ymin = max(ymin, 0)
    xmax = min(xmax, image_width-1)
    ymax = min(ymax, image_height-1)
    anno_x = (anno_x - xmin) / (xmax - xmin)
    anno_y = (anno_y - ymin) / (ymax - ymin)
    anno = np.concatenate([anno_x.reshape(-1,1), anno_y.reshape(-1,1)], axis=1)
    anno = list(anno)
    anno = [list(x) for x in anno]
    image_crop = image[int(ymin):int(ymax), int(xmin):int(xmax), :]
    image_crop = cv2.resize(image_crop, (target_size, target_size))
    return image_crop, anno

def process_wflw(anno, target_size):
    image_name = anno[-1]
    image_path = os.path.join('..', 'data', 'WFLW', 'WFLW_images', image_name)
    image = cv2.imread(image_path)
    image_height, image_width, _ = image.shape
    lms = anno[:196]
    lms = [float(x) for x in lms]
    lms_x = lms[0::2]
    lms_y = lms[1::2]
    lms_x = [x if x >=0 else 0 for x in lms_x] 
    lms_x = [x if x <=image_width else image_width for x in lms_x] 
    lms_y = [y if y >=0 else 0 for y in lms_y] 
    lms_y = [y if y <=image_height else image_height for y in lms_y] 
    lms = [[x,y] for x,y in zip(lms_x, lms_y)]
    lms = [x for z in lms for x in z]
    bbox = anno[196:200]
    bbox = [float(x) for x in bbox]
    attrs = anno[200:206]
    attrs = np.array([int(x) for x in attrs])
    bbox_xmin, bbox_ymin, bbox_xmax, bbox_ymax = bbox

    width = bbox_xmax - bbox_xmin
    height = bbox_ymax - bbox_ymin
    scale = 1.2  
    bbox_xmin -= width * (scale-1)/2
    bbox_ymin -= height * (scale-1)/2
    bbox_xmax += width * (scale-1)/2
    bbox_ymax += height * (scale-1)/2
    bbox_xmin = max(bbox_xmin, 0)
    bbox_ymin = max(bbox_ymin, 0)
    bbox_xmax = min(bbox_xmax, image_width-1)
    bbox_ymax = min(bbox_ymax, image_height-1)
    width = bbox_xmax - bbox_xmin
    height = bbox_ymax - bbox_ymin
    image_crop = image[int(bbox_ymin):int(bbox_ymax), int(bbox_xmin):int(bbox_xmax), :]
    image_crop = cv2.resize(image_crop, (target_size, target_size))

    tmp1 = [bbox_xmin, bbox_ymin]*98
    tmp1 = np.array(tmp1)
    tmp2 = [width, height]*98
    tmp2 = np.array(tmp2)
    lms = np.array(lms) - tmp1
    lms = lms / tmp2
    lms = lms.tolist()
    lms = zip(lms[0::2], lms[1::2])
    return image_crop, list(lms) 

def process_aflw(root_folder, image_name, bbox, anno, target_size):
    image = cv2.imread(os.path.join(root_folder, 'AFLW', 'flickr', image_name))
    image_height, image_width, _ = image.shape
    anno_x = anno[:19]
    anno_y = anno[19:]
    anno_x = [x if x >=0 else 0 for x in anno_x] 
    anno_x = [x if x <=image_width else image_width for x in anno_x] 
    anno_y = [y if y >=0 else 0 for y in anno_y] 
    anno_y = [y if y <=image_height else image_height for y in anno_y] 
    anno_x_min = min(anno_x)
    anno_x_max = max(anno_x)
    anno_y_min = min(anno_y)
    anno_y_max = max(anno_y)
    xmin, xmax, ymin, ymax = bbox
    
    xmin = max(xmin, 0)
    ymin = max(ymin, 0)
    xmax = min(xmax, image_width-1)
    ymax = min(ymax, image_height-1)

    image_crop = image[int(ymin):int(ymax), int(xmin):int(xmax), :]
    image_crop = cv2.resize(image_crop, (target_size, target_size))

    anno_x = (np.array(anno_x) - xmin) / (xmax - xmin)
    anno_y = (np.array(anno_y) - ymin) / (ymax - ymin)

    anno = np.concatenate([anno_x.reshape(-1,1), anno_y.reshape(-1,1)], axis=1).flatten()
    anno = zip(anno[0::2], anno[1::2])
    return image_crop, anno

def gen_meanface(root_folder, data_name):
    with open(os.path.join(root_folder, data_name, 'train.txt'), 'r') as f:
        annos = f.readlines()
    annos = [x.strip().split()[1:] for x in annos]
    annos = [[float(x) for x in anno] for anno in annos]
    annos = np.array(annos)
    meanface = np.mean(annos, axis=0)
    meanface = meanface.tolist()
    meanface = [str(x) for x in meanface]
    
    with open(os.path.join(root_folder, data_name, 'meanface.txt'), 'w') as f:
        f.write(' '.join(meanface))

def convert_wflw(root_folder, data_name):
    with open(os.path.join('../data/WFLW/test.txt'), 'r') as f:
        annos = f.readlines()
    annos = [x.strip().split() for x in annos]
    annos_new = []
    for anno in annos:
        annos_new.append([])
        # name
        annos_new[-1].append(anno[0])
        anno = anno[1:]
        # jaw
        for i in range(17):
            annos_new[-1].append(anno[i*2*2])
            annos_new[-1].append(anno[i*2*2+1])
        # left eyebrow
        annos_new[-1].append(anno[33*2])
        annos_new[-1].append(anno[33*2+1])
        annos_new[-1].append(anno[34*2])
        annos_new[-1].append(str((float(anno[34*2+1])+float(anno[41*2+1]))/2))
        annos_new[-1].append(anno[35*2])
        annos_new[-1].append(str((float(anno[35*2+1])+float(anno[40*2+1]))/2))
        annos_new[-1].append(anno[36*2])
        annos_new[-1].append(str((float(anno[36*2+1])+float(anno[39*2+1]))/2))
        annos_new[-1].append(anno[37*2])
        annos_new[-1].append(str((float(anno[37*2+1])+float(anno[38*2+1]))/2))
        # right eyebrow
        annos_new[-1].append(anno[42*2])
        annos_new[-1].append(str((float(anno[42*2+1])+float(anno[50*2+1]))/2))
        annos_new[-1].append(anno[43*2])
        annos_new[-1].append(str((float(anno[43*2+1])+float(anno[49*2+1]))/2))
        annos_new[-1].append(anno[44*2])
        annos_new[-1].append(str((float(anno[44*2+1])+float(anno[48*2+1]))/2))
        annos_new[-1].append(anno[45*2])
        annos_new[-1].append(str((float(anno[45*2+1])+float(anno[47*2+1]))/2))
        annos_new[-1].append(anno[46*2])
        annos_new[-1].append(anno[46*2+1])
        # nose
        for i in range(51, 60):
            annos_new[-1].append(anno[i*2])
            annos_new[-1].append(anno[i*2+1])
        # left eye
        annos_new[-1].append(anno[60*2])
        annos_new[-1].append(anno[60*2+1])
        annos_new[-1].append(str(0.666*float(anno[61*2])+0.333*float(anno[62*2])))
        annos_new[-1].append(str(0.666*float(anno[61*2+1])+0.333*float(anno[62*2+1])))
        annos_new[-1].append(str(0.666*float(anno[63*2])+0.333*float(anno[62*2])))
        annos_new[-1].append(str(0.666*float(anno[63*2+1])+0.333*float(anno[62*2+1])))
        annos_new[-1].append(anno[64*2])
        annos_new[-1].append(anno[64*2+1])
        annos_new[-1].append(str(0.666*float(anno[65*2])+0.333*float(anno[66*2])))
        annos_new[-1].append(str(0.666*float(anno[65*2+1])+0.333*float(anno[66*2+1])))
        annos_new[-1].append(str(0.666*float(anno[67*2])+0.333*float(anno[66*2])))
        annos_new[-1].append(str(0.666*float(anno[67*2+1])+0.333*float(anno[66*2+1])))
        # right eye
        annos_new[-1].append(anno[68*2])
        annos_new[-1].append(anno[68*2+1])
        annos_new[-1].append(str(0.666*float(anno[69*2])+0.333*float(anno[70*2])))
        annos_new[-1].append(str(0.666*float(anno[69*2+1])+0.333*float(anno[70*2+1])))
        annos_new[-1].append(str(0.666*float(anno[71*2])+0.333*float(anno[70*2])))
        annos_new[-1].append(str(0.666*float(anno[71*2+1])+0.333*float(anno[70*2+1])))
        annos_new[-1].append(anno[72*2])
        annos_new[-1].append(anno[72*2+1])
        annos_new[-1].append(str(0.666*float(anno[73*2])+0.333*float(anno[74*2])))
        annos_new[-1].append(str(0.666*float(anno[73*2+1])+0.333*float(anno[74*2+1])))
        annos_new[-1].append(str(0.666*float(anno[75*2])+0.333*float(anno[74*2])))
        annos_new[-1].append(str(0.666*float(anno[75*2+1])+0.333*float(anno[74*2+1])))
        # mouth
        for i in range(76, 96):
            annos_new[-1].append(anno[i*2])
            annos_new[-1].append(anno[i*2+1])

    with open(os.path.join(root_folder, data_name, 'test.txt'), 'w') as f:
        for anno in annos_new:
            f.write(' '.join(anno)+'\n')


def gen_data(root_folder, data_name, target_size):
    if not os.path.exists(os.path.join(root_folder, data_name, 'images_train')):
        os.mkdir(os.path.join(root_folder, data_name, 'images_train'))
    if not os.path.exists(os.path.join(root_folder, data_name, 'images_test')):
        os.mkdir(os.path.join(root_folder, data_name, 'images_test'))

    ################################################################################################################
    if data_name == 'data_300W':
        folders_train = ['afw', 'helen/trainset', 'lfpw/trainset']
        annos_train = {}
        for folder_train in folders_train:
            all_files = sorted(os.listdir(os.path.join(root_folder, data_name, folder_train)))
            image_files = [x for x in all_files if '.pts' not in x]
            label_files = [x for x in all_files if '.pts' in x]
            assert len(image_files) == len(label_files)
            for image_name, label_name in zip(image_files, label_files):
                print(image_name)
                image_crop, anno = process_300w(os.path.join(root_folder, 'data_300W'), folder_train, image_name, label_name, target_size)
                image_crop_name = folder_train.replace('/', '_')+'_'+image_name
                cv2.imwrite(os.path.join(root_folder, data_name, 'images_train', image_crop_name), image_crop)
                annos_train[image_crop_name] = anno
        with open(os.path.join(root_folder, data_name, 'train.txt'), 'w') as f:
            for image_crop_name, anno in annos_train.items():
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
        

        folders_test = ['helen/testset', 'lfpw/testset', 'ibug']
        annos_test = {}
        for folder_test in folders_test:
            all_files = sorted(os.listdir(os.path.join(root_folder, data_name, folder_test)))
            image_files = [x for x in all_files if '.pts' not in x]
            label_files = [x for x in all_files if '.pts' in x]
            assert len(image_files) == len(label_files)
            for image_name, label_name in zip(image_files, label_files):
                print(image_name)
                image_crop, anno = process_300w(os.path.join(root_folder, 'data_300W'), folder_test, image_name, label_name, target_size)
                image_crop_name = folder_test.replace('/', '_')+'_'+image_name
                cv2.imwrite(os.path.join(root_folder, data_name, 'images_test', image_crop_name), image_crop)
                annos_test[image_crop_name] = anno
        with open(os.path.join(root_folder, data_name, 'test.txt'), 'w') as f:
            for image_crop_name, anno in annos_test.items():
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')

        annos = None
        with open(os.path.join(root_folder, data_name, 'test.txt'), 'r') as f:
            annos = f.readlines()
        with open(os.path.join(root_folder, data_name, 'test_common.txt'), 'w') as f:
            for anno in annos:
                if not 'ibug' in anno:
                    f.write(anno)
        with open(os.path.join(root_folder, data_name, 'test_challenge.txt'), 'w') as f:
            for anno in annos:
                if 'ibug' in anno:
                    f.write(anno)

        gen_meanface(root_folder, data_name)
    ################################################################################################################
    elif data_name == 'COFW':
        train_file = 'COFW_train_color.mat'
        train_mat = hdf5storage.loadmat(os.path.join(root_folder, 'COFW', train_file))
        images = train_mat['IsTr']
        bboxes = train_mat['bboxesTr']
        annos = train_mat['phisTr']

        count = 1
        with open(os.path.join(root_folder, 'COFW', 'train.txt'), 'w') as f:
            for i in range(images.shape[0]):
                image = images[i, 0]
                # grayscale
                if len(image.shape) == 2:
                    image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
                # swap rgb channel to bgr
                else:
                    image = image[:,:,::-1]
                bbox = bboxes[i, :]
                anno = annos[i, :]
                image_crop, anno = process_cofw(image, bbox, anno, target_size)
                pad_num = 4-len(str(count))
                image_crop_name = 'cofw_train_' + '0' * pad_num + str(count) + '.jpg'
                print(image_crop_name)
                cv2.imwrite(os.path.join(root_folder, 'COFW', 'images_train', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
                count += 1

        test_file = 'COFW_test_color.mat'
        test_mat = hdf5storage.loadmat(os.path.join(root_folder, 'COFW', test_file))
        images = test_mat['IsT']
        bboxes = test_mat['bboxesT']
        annos = test_mat['phisT']

        count = 1
        with open(os.path.join(root_folder, 'COFW', 'test.txt'), 'w') as f:
            for i in range(images.shape[0]):
                image = images[i, 0]
                # grayscale
                if len(image.shape) == 2:
                    image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
                # swap rgb channel to bgr
                else:
                    image = image[:,:,::-1]
                bbox = bboxes[i, :]
                anno = annos[i, :]
                image_crop, anno = process_cofw(image, bbox, anno, target_size)
                pad_num = 4-len(str(count))
                image_crop_name = 'cofw_test_' + '0' * pad_num + str(count) + '.jpg'
                print(image_crop_name)
                cv2.imwrite(os.path.join(root_folder, 'COFW', 'images_test', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
                count += 1
        gen_meanface(root_folder, data_name)
    ################################################################################################################
    elif data_name == 'WFLW':
        train_file = 'list_98pt_rect_attr_train.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_rect_attr_train_test', train_file), 'r') as f:
            annos_train = f.readlines()
        annos_train = [x.strip().split() for x in annos_train]
        count = 1
        with open(os.path.join(root_folder, 'WFLW', 'train.txt'), 'w') as f:
            for anno_train in annos_train:
                image_crop, anno = process_wflw(anno_train, target_size)
                pad_num = 4-len(str(count))
                image_crop_name = 'wflw_train_' + '0' * pad_num + str(count) + '.jpg'
                print(image_crop_name)
                cv2.imwrite(os.path.join(root_folder, 'WFLW', 'images_train', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
                count += 1

        test_file = 'list_98pt_rect_attr_test.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_rect_attr_train_test', test_file), 'r') as f:
            annos_test = f.readlines()
        annos_test = [x.strip().split() for x in annos_test]
        names_mapping = {}
        count = 1
        with open(os.path.join(root_folder, 'WFLW', 'test.txt'), 'w') as f:
            for anno_test in annos_test:
                image_crop, anno = process_wflw(anno_test, target_size)
                pad_num = 4-len(str(count))
                image_crop_name = 'wflw_test_' + '0' * pad_num + str(count) + '.jpg'
                print(image_crop_name)
                names_mapping[anno_test[0]+'_'+anno_test[-1]] = [image_crop_name, anno]
                cv2.imwrite(os.path.join(root_folder, 'WFLW', 'images_test', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in list(anno):
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
                count += 1

        test_pose_file = 'list_98pt_test_largepose.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_pose_file), 'r') as f:
            annos_pose_test = f.readlines()
        names_pose = [x.strip().split() for x in annos_pose_test]
        names_pose = [x[0]+'_'+x[-1] for x in names_pose]
        with open(os.path.join(root_folder, 'WFLW', 'test_pose.txt'), 'w') as f:
            for name_pose in names_pose:
                if name_pose in names_mapping:
                    image_crop_name, anno = names_mapping[name_pose]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)

        test_expr_file = 'list_98pt_test_expression.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_expr_file), 'r') as f:
            annos_expr_test = f.readlines()
        names_expr = [x.strip().split() for x in annos_expr_test]
        names_expr = [x[0]+'_'+x[-1] for x in names_expr]
        with open(os.path.join(root_folder, 'WFLW', 'test_expr.txt'), 'w') as f:
            for name_expr in names_expr:
                if name_expr in names_mapping:
                    image_crop_name, anno = names_mapping[name_expr]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)

        test_illu_file = 'list_98pt_test_illumination.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_illu_file), 'r') as f:
            annos_illu_test = f.readlines()
        names_illu = [x.strip().split() for x in annos_illu_test]
        names_illu = [x[0]+'_'+x[-1] for x in names_illu]
        with open(os.path.join(root_folder, 'WFLW', 'test_illu.txt'), 'w') as f:
            for name_illu in names_illu:
                if name_illu in names_mapping:
                    image_crop_name, anno = names_mapping[name_illu]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)

        test_mu_file = 'list_98pt_test_makeup.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_mu_file), 'r') as f:
            annos_mu_test = f.readlines()
        names_mu = [x.strip().split() for x in annos_mu_test]
        names_mu = [x[0]+'_'+x[-1] for x in names_mu]
        with open(os.path.join(root_folder, 'WFLW', 'test_mu.txt'), 'w') as f:
            for name_mu in names_mu:
                if name_mu in names_mapping:
                    image_crop_name, anno = names_mapping[name_mu]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)

        test_occu_file = 'list_98pt_test_occlusion.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_occu_file), 'r') as f:
            annos_occu_test = f.readlines()
        names_occu = [x.strip().split() for x in annos_occu_test]
        names_occu = [x[0]+'_'+x[-1] for x in names_occu]
        with open(os.path.join(root_folder, 'WFLW', 'test_occu.txt'), 'w') as f:
            for name_occu in names_occu:
                if name_occu in names_mapping:
                    image_crop_name, anno = names_mapping[name_occu]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)


        test_blur_file = 'list_98pt_test_blur.txt'
        with open(os.path.join(root_folder, 'WFLW', 'WFLW_annotations', 'list_98pt_test', test_blur_file), 'r') as f:
            annos_blur_test = f.readlines()
        names_blur = [x.strip().split() for x in annos_blur_test]
        names_blur = [x[0]+'_'+x[-1] for x in names_blur]
        with open(os.path.join(root_folder, 'WFLW', 'test_blur.txt'), 'w') as f:
            for name_blur in names_blur:
                if name_blur in names_mapping:
                    image_crop_name, anno = names_mapping[name_blur]
                    f.write(image_crop_name+' ')
                    for x,y in anno:
                        f.write(str(x)+' '+str(y)+' ')
                    f.write('\n')
                else:
                    print('error!')
                    exit(0)
        gen_meanface(root_folder, data_name)
    ################################################################################################################
    elif data_name == 'AFLW':
        mat = hdf5storage.loadmat('../data/AFLW/AFLWinfo_release.mat')
        bboxes = mat['bbox']
        annos = mat['data']
        mask_new = mat['mask_new']
        nameList = mat['nameList']
        ra = mat['ra'][0]
        train_indices = ra[:20000]
        test_indices = ra[20000:]

        with open(os.path.join(root_folder, 'AFLW', 'train.txt'), 'w') as f:
            for index in train_indices:
                # from matlab index
                image_name = nameList[index-1][0][0]
                bbox = bboxes[index-1]
                anno = annos[index-1]
                image_crop, anno = process_aflw(root_folder, image_name, bbox, anno, target_size)
                pad_num = 5-len(str(index))
                image_crop_name = 'aflw_train_' + '0' * pad_num + str(index) + '.jpg'
                print(image_crop_name)
                cv2.imwrite(os.path.join(root_folder, 'AFLW', 'images_train', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')

        with open(os.path.join(root_folder, 'AFLW', 'test.txt'), 'w') as f:
            for index in test_indices:
                # from matlab index
                image_name = nameList[index-1][0][0]
                bbox = bboxes[index-1]
                anno = annos[index-1]
                image_crop, anno = process_aflw(root_folder, image_name, bbox, anno, target_size)
                pad_num = 5-len(str(index))
                image_crop_name = 'aflw_test_' + '0' * pad_num + str(index) + '.jpg'
                print(image_crop_name)
                cv2.imwrite(os.path.join(root_folder, 'AFLW', 'images_test', image_crop_name), image_crop)
                f.write(image_crop_name+' ')
                for x,y in anno:
                    f.write(str(x)+' '+str(y)+' ')
                f.write('\n')
        gen_meanface(root_folder, data_name)
    ################################################################################################################
    elif data_name == 'LaPa':
        pass
        # TODO
    else:
        print('Wrong data!')

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('please input the data name.')
        print('1. data_300W')
        print('2. COFW')
        print('3. WFLW')
        print('4. AFLW')
        print('5. LaPa')
        exit(0)
    else:
        data_name = sys.argv[1]
        gen_data('../data', data_name, 256)





import cv2
from matplotlib import pyplot as plt

from base import VisionTestCase
from visage.bounding_boxes.face_boxes_v2 import FaceBoxesV2


class FaceBoxTest(VisionTestCase):

    def test_face_box(self):
        img = self._get_obama_image(
            normalize=False,
            channel_first=False,
            keep_aspect_ratio=True,
            resize=None)
        img = (img * 255).int().numpy()

        detector = FaceBoxesV2()
        # detector = detector.cuda()
        my_thresh = 0.6

        detections = detector.detect(img, my_thresh)
        print(detections)

        cv2.rectangle(img,
                      detections[0].get_point1(),
                      detections[0].get_point2(),
                      (255, 0, 0),
                      10)

        plt.imshow(img)
        plt.show()



# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------
import inspect
from pathlib import Path

try:
    from .nms.cpu_nms import cpu_nms
except ModuleNotFoundError:
    # Hacky way to call the build script that creates the cpu_nms python files
    # Ideally, this would be a post-install script, but there does not seem to be a good way
    import os
    import subprocess
    import visage

    cwd = os.getcwd()
    new_wd = Path(inspect.getfile(visage)).parent.parent#.parent
    cmd = ["python", "face_parser/modules/FaceBoxesV2/utils/build.py", "build_ext", "--inplace"]
    os.chdir(new_wd)
    print(f"Running: {' '.join(cmd)}")
    print(f"From directory: {new_wd}")
    # subprocess.call(f"python {Path(__file__).parent}/build.py build_ext --inplace")
    subprocess.call(cmd)
    os.chdir(cwd)

    from .nms.cpu_nms import cpu_nms


def nms(dets, thresh):
    """Dispatch to either CPU or GPU NMS implementations."""

    if dets.shape[0] == 0:
        return []
    return cpu_nms(dets, thresh)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, List

import numpy as np
import requests
import torch
from torch import nn

from visage.modules.FaceBoxesV2.detector import Detector
from visage.modules.FaceBoxesV2.utils.box_utils import decode
from visage.modules.FaceBoxesV2.utils.config import cfg
from visage.modules.FaceBoxesV2.utils.faceboxes import FaceBoxesV2Internal
from visage.modules.FaceBoxesV2.utils.nms_wrapper import nms
from visage.modules.FaceBoxesV2.utils.prior_box import PriorBox


@dataclass
class DetectedBBox:
    x: int
    y: int
    width: int
    height: int
    score: float

    def get_point1(self) -> Tuple[int, int]:
        return self.x, self.y

    def get_point2(self) -> Tuple[int, int]:
        return self.x + self.width, self.y + self.height


class FaceBoxesDetector(Detector):
    def __init__(self):
        ckpt_folder = f"{Path().home()}/.cache/torch/face-parser"
        ckpt_path = f"{ckpt_folder}/FaceBoxesV2.pth"
        os.makedirs(ckpt_folder, exist_ok=True)
        # Download path from PIPNet repository:
        # https://github.com/jhb86253817/PIPNet/blob/master/FaceBoxesV2/weights/FaceBoxesV2.pth
        download_path = "https://github.com/jhb86253817/PIPNet/raw/master/FaceBoxesV2/weights/FaceBoxesV2.pth"

        if not Path(ckpt_path).exists():
            print(f"Downloading {download_path} into {ckpt_path}")
            request = requests.get(download_path, allow_redirects=True)
            open(ckpt_path, 'wb').write(request.content)

        super().__init__('FaceBoxes', ckpt_path)

        self.name = 'FaceBoxesDetector'
        self.net = FaceBoxesV2Internal(phase='test', size=None, num_classes=2)  # initialize detector
        state_dict = torch.load(self.model_weights, map_location=torch.device('cpu'))
        # create new OrderedDict that does not contain `module.`
        from collections import OrderedDict
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            name = k[7:]  # remove `module.`
            new_state_dict[name] = v
        # load params
        self.net.load_state_dict(new_state_dict)
        self.net.eval()

        self.device_indicator_param = nn.Parameter()

    def detect(self, image: np.ndarray, thresh: float = 0.6) -> List[DetectedBBox]:

        device = self.device_indicator_param
        image_scale = image

        scale = torch.Tensor([image_scale.shape[1], image_scale.shape[0], image_scale.shape[1], image_scale.shape[0]])
        image_scale = torch.from_numpy(image_scale.transpose(2, 0, 1)).to(device).int()
        mean_tmp = torch.IntTensor([104, 117, 123]).to(device).int()
        mean_tmp = mean_tmp.unsqueeze(1).unsqueeze(2)
        image_scale -= mean_tmp
        image_scale = image_scale.float().unsqueeze(0)
        scale = scale.to(device)

        with torch.no_grad():
            out = self.net(image_scale)
            # priorbox = PriorBox(cfg, out[2], (image_scale.size()[2], image_scale.size()[3]), phase='test')
            priorbox = PriorBox(cfg, image_size=(image_scale.size()[2], image_scale.size()[3]))
            priors = priorbox.forward()
            priors = priors.to(device)
            loc, conf = out
            prior_data = priors.data
            boxes = decode(loc.data.squeeze(0), prior_data, cfg['variance'])
            boxes = boxes * scale
            boxes = boxes.cpu().numpy()
            scores = conf.data.cpu().numpy()[:, 1]

            # ignore low scores
            inds = np.where(scores > thresh)[0]
            boxes = boxes[inds]
            scores = scores[inds]

            # keep top-K before NMS
            order = scores.argsort()[::-1][:5000]
            boxes = boxes[order]
            scores = scores[order]

            # do NMS
            dets = np.hstack((boxes, scores[:, np.newaxis])).astype(np.float32, copy=False)
            keep = nms(dets, 0.3)
            dets = dets[keep, :]

            dets = dets[:750, :]
            detections_scale = []
            for i in range(dets.shape[0]):
                xmin = int(dets[i][0])
                ymin = int(dets[i][1])
                xmax = int(dets[i][2])
                ymax = int(dets[i][3])
                score = dets[i][4]
                width = xmax - xmin
                height = ymax - ymin
                detections_scale.append(DetectedBBox(xmin, ymin, width, height, score))

        return detections_scale



import numpy as np
import torch
from dreifus.image import Img
from insightface.app import FaceAnalysis
from torch import cosine_similarity


class CSIMEvaluator:

    def __init__(self):
        self._app = FaceAnalysis(allowed_modules=['detection', 'recognition'])
        self._app.prepare(ctx_id=0, det_size=(512, 512))

    def __call__(self, predictions: torch.Tensor, references: torch.Tensor) -> torch.Tensor:
        embeddings_prediction = []
        embeddings_reference = []
        for prediction, reference, in zip(predictions, references):
            prediction = Img.from_torch(prediction).to_numpy().img
            reference = Img.from_torch(reference).to_numpy().img

            faces_prediction = self._app.get(prediction)
            faces_reference = self._app.get(reference)

            if len(faces_prediction) > 0 and len(faces_reference) > 0:
                embedding_prediction = faces_prediction[0]['embedding']
                embedding_reference = faces_reference[0]['embedding']

                embeddings_prediction.append(embedding_prediction)
                embeddings_reference.append(embedding_reference)

        if len(embeddings_prediction) > 0:
            embeddings_prediction = np.stack(embeddings_prediction)
            embeddings_reference = np.stack(embeddings_reference)
            csim = cosine_similarity(torch.from_numpy(embeddings_prediction), torch.from_numpy(embeddings_reference)).mean()
        else:
            csim = torch.tensor(-1)

        return csim

class UnpairedCSIMEvaluator:

    def __init__(self):
        self._app = FaceAnalysis(allowed_modules=['detection', 'recognition'])
        self._app.prepare(ctx_id=0, det_size=(512, 512))

    def __call__(self, predictions: torch.Tensor, references: torch.Tensor) -> torch.Tensor:
        embeddings_prediction = []
        embeddings_reference = []
        for prediction in predictions:
            prediction = Img.from_torch(prediction).to_numpy().img

            faces_prediction = self._app.get(prediction)

            if len(faces_prediction):
                embedding_prediction = faces_prediction[0]['embedding']

                embeddings_prediction.append(embedding_prediction)

        for reference in references:
            reference = Img.from_torch(reference).to_numpy().img

            faces_reference = self._app.get(reference)

            if len(faces_reference) > 0:
                embedding_reference = faces_reference[0]['embedding']

                embeddings_reference.append(embedding_reference)

        embedding_prediction = np.stack(embeddings_prediction).mean(axis=0)
        embedding_reference = np.stack(embeddings_reference).mean(axis=0)
        csim = cosine_similarity(torch.from_numpy(embedding_prediction)[None], torch.from_numpy(embedding_reference)[None])

        return csim



import inspect
from pathlib import Path

import cv2
import numpy as np
import torch
import torchvision.models as models
from PIL import Image
from torch import nn
from torchvision.transforms import transforms

import visage.modules.PIPNet
from visage.landmark_detection.pipnet_config import PIPNetConfig, PIPNet_WFLW_r18_config
from visage.modules.FaceBoxesV2.faceboxes_detector import DetectedBBox
from visage.modules.PIPNet.lib.functions import get_meanface, forward_pip
from visage.modules.PIPNet.lib.mobilenetv3 import mobilenetv3_large
from visage.modules.PIPNet.lib.networks import Pip_resnet18, Pip_resnet50, Pip_resnet101, Pip_mbnetv2, Pip_mbnetv3


class PIPNet(nn.Module):

    def __init__(self, config: PIPNetConfig = PIPNet_WFLW_r18_config):
        super(PIPNet, self).__init__()

        self._cfg = config
        self._det_box_scale = 1.2

        if self._cfg.backbone == 'resnet18':
            resnet18 = models.resnet18(pretrained=self._cfg.pretrained)
            net = Pip_resnet18(resnet18, self._cfg.num_nb, num_lms=self._cfg.num_lms, input_size=self._cfg.input_size,
                               net_stride=self._cfg.net_stride)
        elif self._cfg.backbone == 'resnet50':
            resnet50 = models.resnet50(pretrained=self._cfg.pretrained)
            net = Pip_resnet50(resnet50, self._cfg.num_nb, num_lms=self._cfg.num_lms, input_size=self._cfg.input_size,
                               net_stride=self._cfg.net_stride)
        elif self._cfg.backbone == 'resnet101':
            resnet101 = models.resnet101(pretrained=self._cfg.pretrained)
            net = Pip_resnet101(resnet101, self._cfg.num_nb, num_lms=self._cfg.num_lms, input_size=self._cfg.input_size,
                                net_stride=self._cfg.net_stride)
        elif self._cfg.backbone == 'mobilenet_v2':
            mbnet = models.mobilenet_v2(pretrained=self._cfg.pretrained)
            net = Pip_mbnetv2(mbnet, self._cfg.num_nb, num_lms=self._cfg.num_lms, input_size=self._cfg.input_size,
                              net_stride=self._cfg.net_stride)
        elif self._cfg.backbone == 'mobilenet_v3':
            mbnet = mobilenetv3_large()
            if self._cfg.pretrained:
                mbnet.load_state_dict(torch.load('lib/mobilenetv3-large-1cd25616.pth'))
            net = Pip_mbnetv3(mbnet, self._cfg.num_nb, num_lms=self._cfg.num_lms, input_size=self._cfg.input_size,
                              net_stride=self._cfg.net_stride)
        else:
            raise ValueError(f'Invalid backbone: {self._cfg.backbone}')

        weight_file = f"{Path().home()}/.cache/torch/face-parser/PIPNet/{self._cfg.checkpoint_path}"
        if not Path(weight_file).exists():
            raise ValueError(f"Download {self._cfg.checkpoint_path} from "
                             f"https://drive.google.com/drive/folders/1fz6UQR2TjGvQr4birwqVXusPp6tMAXxq and put it into "
                             f"{Path().home()}/.cache/torch/face-parser/PIPNet/")

        state_dict = torch.load(weight_file, map_location=torch.device('cpu'))
        net.load_state_dict(state_dict)

        self._normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                               std=[0.229, 0.224, 0.225])
        self._preprocess = transforms.Compose(
            [transforms.Resize((self._cfg.input_size, self._cfg.input_size)), transforms.ToTensor(), self._normalize])

        meanface_path = f"{Path(inspect.getfile(visage.modules.PIPNet)).parent}/meanfaces/{self._cfg.meanface_path}"
        meanface_indices, reverse_index1, reverse_index2, max_len = get_meanface(meanface_path, self._cfg.num_nb)

        self._meanface_indices = meanface_indices
        self._reverse_index1 = reverse_index1
        self._reverse_index2 = reverse_index2
        self._max_len = max_len

        net.eval()
        self._net = net
        self._device_indicator_param = nn.Parameter()

    def forward(self, image: np.ndarray, bbox: DetectedBBox) -> np.ndarray:
        device = self._device_indicator_param.device

        if bbox.score == -1:
            landmarks = None
        else:
            image_height, image_width, _ = image.shape
            det_xmin = int(bbox.x)
            det_ymin = int(bbox.y)
            det_width = int(bbox.width)
            det_height = int(bbox.height)
            det_xmax = det_xmin + det_width - 1
            det_ymax = det_ymin + det_height - 1

            det_xmin -= int(det_width * (self._det_box_scale - 1) / 2)
            # remove a part of top area for alignment, see paper for details
            det_ymin += int(det_height * (self._det_box_scale - 1) / 2)
            det_xmax += int(det_width * (self._det_box_scale - 1) / 2)
            det_ymax += int(det_height * (self._det_box_scale - 1) / 2)
            det_xmin = max(det_xmin, 0)
            det_ymin = max(det_ymin, 0)
            det_xmax = min(det_xmax, image_width - 1)
            det_ymax = min(det_ymax, image_height - 1)
            det_width = det_xmax - det_xmin + 1
            det_height = det_ymax - det_ymin + 1
            # cv2.rectangle(image, (det_xmin, det_ymin), (det_xmax, det_ymax), (0, 0, 255), 2)
            det_crop = image[det_ymin:det_ymax, det_xmin:det_xmax, :].astype(float)
            # plt.imshow(det_crop)
            # plt.show()
            det_crop = cv2.resize(det_crop, (self._cfg.input_size, self._cfg.input_size))
            inputs = Image.fromarray(det_crop[:, :, ::-1].astype('uint8'), 'RGB')
            inputs = self._preprocess(inputs).unsqueeze(0)
            inputs = inputs.to(device)
            lms_pred_x, lms_pred_y, lms_pred_nb_x, lms_pred_nb_y, outputs_cls, max_cls = forward_pip(
                self._net,
                inputs,
                self._preprocess,
                self._cfg.input_size,
                self._cfg.net_stride,
                self._cfg.num_nb)
            tmp_nb_x = lms_pred_nb_x[self._reverse_index1, self._reverse_index2].view(self._cfg.num_lms, self._max_len)
            tmp_nb_y = lms_pred_nb_y[self._reverse_index1, self._reverse_index2].view(self._cfg.num_lms, self._max_len)
            tmp_x = torch.mean(torch.cat((lms_pred_x, tmp_nb_x), dim=1), dim=1).view(-1, 1)
            tmp_y = torch.mean(torch.cat((lms_pred_y, tmp_nb_y), dim=1), dim=1).view(-1, 1)
            lms_pred_merge = torch.cat((tmp_x, tmp_y), dim=1).flatten()
            lms_pred_merge = lms_pred_merge.cpu().numpy()
            pred_export = np.zeros([self._cfg.num_lms, 2])
            for ii in range(self._cfg.num_lms):
                x_pred = lms_pred_merge[ii * 2] * det_width
                y_pred = lms_pred_merge[ii * 2 + 1] * det_height
                pred_export[ii, 0] = x_pred + det_xmin
                pred_export[ii, 1] = y_pred + det_ymin
            landmarks = pred_export

        return landmarks



#!/usr/bin/python
# -*- encoding: utf-8 -*-
from pathlib import Path
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torchvision import transforms

from .resnet import Resnet18


# from modules.bn import InPlaceABNSync as BatchNorm2d

class BiSeNetFaceParser:

    def __init__(self):
        ckpt_path = f"{Path().home()}/.cache/torch/face-parser/bisenet.pth"
        download_path = "https://drive.google.com/open?id=154JgKpzCPW82qINcVieuPH3fZ2e0P812"

        if not Path(ckpt_path).exists():
            raise ValueError(
                f"No pre-trained BiSeNet found. Download the checkpoint from {download_path} and put it into {ckpt_path}")

        n_classes = 19
        self._model = BiSeNet(n_classes=n_classes).cuda()
        self._model.load_state_dict(torch.load(ckpt_path))
        self._model.eval()

        self._transform = transforms.Compose([
            # transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])

    def parse(self, images: torch.Tensor) -> torch.Tensor:
        """

        :param images:
            Can be one or many images. Either in channel-first or channel-last format:
             - [B, 3, H, W]
             - [B, H, W, 3]
             - [3, H, W]
             - [H, W, 3]
            Images should be normalized to [0, 1]!
        :return:
            A segmentation mask of the same shape as the input, but the channel dimension collapsed
            which indicates 1 out of 20 classes for each pixel.
            Example input: [B, 3, H, W] -> [B, H, W]
        """

        needs_squeeze = False
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
            needs_squeeze = True

        if images.shape[-1] == 3:
            # torchvision transforms requires images in Channel-first format
            images = images.permute(0, 3, 1, 2)

        images = self._transform(images)
        images = images.cuda()
        segmentation_masks = self._model(images)[0]
        segmentation_masks = segmentation_masks.cpu().detach().argmax(1)

        if needs_squeeze:
            segmentation_masks = segmentation_masks.squeeze(0)

        return segmentation_masks

    @classmethod
    def get_label_mapping(cls) -> Dict[int, str]:
        return {0: 'background',
                1: 'skin',
                2: 'l_brow',
                3: 'r_brow',
                4: 'l_eye',
                5: 'r_eye',
                6: 'eye_g',
                7: 'l_ear',
                8: 'r_ear',
                9: 'ear_r',
                10: 'nose',
                11: 'mouth',
                12: 'u_lip',
                13: 'l_lip',
                14: 'neck',
                15: 'neck_l',
                16: 'cloth',
                17: 'hair',
                18: 'hat'}


class ConvBNReLU(nn.Module):
    def __init__(self, in_chan, out_chan, ks=3, stride=1, padding=1, *args, **kwargs):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(in_chan,
                              out_chan,
                              kernel_size=ks,
                              stride=stride,
                              padding=padding,
                              bias=False)
        self.bn = nn.BatchNorm2d(out_chan)
        self.init_weight()

    def forward(self, x):
        x = self.conv(x)
        x = F.relu(self.bn(x))
        return x

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)


class BiSeNetOutput(nn.Module):
    def __init__(self, in_chan, mid_chan, n_classes, *args, **kwargs):
        super(BiSeNetOutput, self).__init__()
        self.conv = ConvBNReLU(in_chan, mid_chan, ks=3, stride=1, padding=1)
        self.conv_out = nn.Conv2d(mid_chan, n_classes, kernel_size=1, bias=False)
        self.init_weight()

    def forward(self, x):
        x = self.conv(x)
        x = self.conv_out(x)
        return x

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params = [], []
        for name, module in self.named_modules():
            if isinstance(module, nn.Linear) or isinstance(module, nn.Conv2d):
                wd_params.append(module.weight)
                if not module.bias is None:
                    nowd_params.append(module.bias)
            elif isinstance(module, nn.BatchNorm2d):
                nowd_params += list(module.parameters())
        return wd_params, nowd_params


class AttentionRefinementModule(nn.Module):
    def __init__(self, in_chan, out_chan, *args, **kwargs):
        super(AttentionRefinementModule, self).__init__()
        self.conv = ConvBNReLU(in_chan, out_chan, ks=3, stride=1, padding=1)
        self.conv_atten = nn.Conv2d(out_chan, out_chan, kernel_size=1, bias=False)
        self.bn_atten = nn.BatchNorm2d(out_chan)
        self.sigmoid_atten = nn.Sigmoid()
        self.init_weight()

    def forward(self, x):
        feat = self.conv(x)
        atten = F.avg_pool2d(feat, feat.size()[2:])
        atten = self.conv_atten(atten)
        atten = self.bn_atten(atten)
        atten = self.sigmoid_atten(atten)
        out = torch.mul(feat, atten)
        return out

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)


class ContextPath(nn.Module):
    def __init__(self, *args, **kwargs):
        super(ContextPath, self).__init__()
        self.resnet = Resnet18()
        self.arm16 = AttentionRefinementModule(256, 128)
        self.arm32 = AttentionRefinementModule(512, 128)
        self.conv_head32 = ConvBNReLU(128, 128, ks=3, stride=1, padding=1)
        self.conv_head16 = ConvBNReLU(128, 128, ks=3, stride=1, padding=1)
        self.conv_avg = ConvBNReLU(512, 128, ks=1, stride=1, padding=0)

        self.init_weight()

    def forward(self, x):
        H0, W0 = x.size()[2:]
        feat8, feat16, feat32 = self.resnet(x)
        H8, W8 = feat8.size()[2:]
        H16, W16 = feat16.size()[2:]
        H32, W32 = feat32.size()[2:]

        avg = F.avg_pool2d(feat32, feat32.size()[2:])
        avg = self.conv_avg(avg)
        avg_up = F.interpolate(avg, (H32, W32), mode='nearest')

        feat32_arm = self.arm32(feat32)
        feat32_sum = feat32_arm + avg_up
        feat32_up = F.interpolate(feat32_sum, (H16, W16), mode='nearest')
        feat32_up = self.conv_head32(feat32_up)

        feat16_arm = self.arm16(feat16)
        feat16_sum = feat16_arm + feat32_up
        feat16_up = F.interpolate(feat16_sum, (H8, W8), mode='nearest')
        feat16_up = self.conv_head16(feat16_up)

        return feat8, feat16_up, feat32_up  # x8, x8, x16

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params = [], []
        for name, module in self.named_modules():
            if isinstance(module, (nn.Linear, nn.Conv2d)):
                wd_params.append(module.weight)
                if not module.bias is None:
                    nowd_params.append(module.bias)
            elif isinstance(module, nn.BatchNorm2d):
                nowd_params += list(module.parameters())
        return wd_params, nowd_params


### This is not used, since I replace this with the resnet feature with the same size
class SpatialPath(nn.Module):
    def __init__(self, *args, **kwargs):
        super(SpatialPath, self).__init__()
        self.conv1 = ConvBNReLU(3, 64, ks=7, stride=2, padding=3)
        self.conv2 = ConvBNReLU(64, 64, ks=3, stride=2, padding=1)
        self.conv3 = ConvBNReLU(64, 64, ks=3, stride=2, padding=1)
        self.conv_out = ConvBNReLU(64, 128, ks=1, stride=1, padding=0)
        self.init_weight()

    def forward(self, x):
        feat = self.conv1(x)
        feat = self.conv2(feat)
        feat = self.conv3(feat)
        feat = self.conv_out(feat)
        return feat

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params = [], []
        for name, module in self.named_modules():
            if isinstance(module, nn.Linear) or isinstance(module, nn.Conv2d):
                wd_params.append(module.weight)
                if not module.bias is None:
                    nowd_params.append(module.bias)
            elif isinstance(module, nn.BatchNorm2d):
                nowd_params += list(module.parameters())
        return wd_params, nowd_params


class FeatureFusionModule(nn.Module):
    def __init__(self, in_chan, out_chan, *args, **kwargs):
        super(FeatureFusionModule, self).__init__()
        self.convblk = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)
        self.conv1 = nn.Conv2d(out_chan,
                               out_chan // 4,
                               kernel_size=1,
                               stride=1,
                               padding=0,
                               bias=False)
        self.conv2 = nn.Conv2d(out_chan // 4,
                               out_chan,
                               kernel_size=1,
                               stride=1,
                               padding=0,
                               bias=False)
        self.relu = nn.ReLU(inplace=True)
        self.sigmoid = nn.Sigmoid()
        self.init_weight()

    def forward(self, fsp, fcp):
        fcat = torch.cat([fsp, fcp], dim=1)
        feat = self.convblk(fcat)
        atten = F.avg_pool2d(feat, feat.size()[2:])
        atten = self.conv1(atten)
        atten = self.relu(atten)
        atten = self.conv2(atten)
        atten = self.sigmoid(atten)
        feat_atten = torch.mul(feat, atten)
        feat_out = feat_atten + feat
        return feat_out

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params = [], []
        for name, module in self.named_modules():
            if isinstance(module, nn.Linear) or isinstance(module, nn.Conv2d):
                wd_params.append(module.weight)
                if not module.bias is None:
                    nowd_params.append(module.bias)
            elif isinstance(module, nn.BatchNorm2d):
                nowd_params += list(module.parameters())
        return wd_params, nowd_params


class BiSeNet(nn.Module):
    def __init__(self, n_classes, *args, **kwargs):
        super(BiSeNet, self).__init__()
        self.cp = ContextPath()
        ## here self.sp is deleted
        self.ffm = FeatureFusionModule(256, 256)
        self.conv_out = BiSeNetOutput(256, 256, n_classes)
        self.conv_out16 = BiSeNetOutput(128, 64, n_classes)
        self.conv_out32 = BiSeNetOutput(128, 64, n_classes)
        self.init_weight()

    def forward(self, x):
        H, W = x.size()[2:]
        feat_res8, feat_cp8, feat_cp16 = self.cp(x)  # here return res3b1 feature
        feat_sp = feat_res8  # use res3b1 feature to replace spatial path feature
        feat_fuse = self.ffm(feat_sp, feat_cp8)

        feat_out = self.conv_out(feat_fuse)
        feat_out16 = self.conv_out16(feat_cp8)
        feat_out32 = self.conv_out32(feat_cp16)

        feat_out = F.interpolate(feat_out, (H, W), mode='bilinear', align_corners=True)
        feat_out16 = F.interpolate(feat_out16, (H, W), mode='bilinear', align_corners=True)
        feat_out32 = F.interpolate(feat_out32, (H, W), mode='bilinear', align_corners=True)
        return feat_out, feat_out16, feat_out32

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None: nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params, lr_mul_wd_params, lr_mul_nowd_params = [], [], [], []
        for name, child in self.named_children():
            child_wd_params, child_nowd_params = child.get_params()
            if isinstance(child, FeatureFusionModule) or isinstance(child, BiSeNetOutput):
                lr_mul_wd_params += child_wd_params
                lr_mul_nowd_params += child_nowd_params
            else:
                wd_params += child_wd_params
                nowd_params += child_nowd_params
        return wd_params, nowd_params, lr_mul_wd_params, lr_mul_nowd_params


if __name__ == "__main__":
    net = BiSeNet(19)
    net.cuda()
    net.eval()
    in_ten = torch.randn(16, 3, 640, 480).cuda()
    out, out16, out32 = net(in_ten)
    print(out.shape)

    net.get_params()



import torch
from elias.util.batch import batchify_sliced
from torch import nn
from torchmetrics.image import LearnedPerceptualImagePatchSimilarity


class LPIPSEvaluator(nn.Module):

    def __init__(self, batch_size: int = 16):
        super(LPIPSEvaluator, self).__init__()
        self._batch_size = batch_size
        self._lpips_evaluator = LearnedPerceptualImagePatchSimilarity(normalize=True)

    def __call__(self, predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        lpipses = []
        for predictions_batch, targets_batch in zip(batchify_sliced(predictions, self._batch_size), batchify_sliced(targets, self._batch_size)):
            lpipses.append(self._lpips_evaluator(predictions_batch, targets_batch))

        lpips = torch.stack(lpipses).mean()
        return lpips



import cv2
from matplotlib import pyplot as plt

from base import VisionTestCase
from visage.bounding_boxes.face_boxes_v2 import FaceBoxesV2
from visage.landmark_detection.pipnet import PIPNet


class PIPNetTest(VisionTestCase):

    def test_pipnet(self):
        pip_net = PIPNet()
        pip_net = pip_net.cuda()

        facebox_detector = FaceBoxesV2()
        facebox_detector = facebox_detector.cuda()

        img = self._get_obama_image(
            normalize=False,
            channel_first=False,
            keep_aspect_ratio=True,
            resize=None)
        img = (img * 255).int().numpy()

        detected_bboxes = facebox_detector.detect(img)
        landmarks = pip_net.forward(img, detected_bboxes[0])

        for x, y in landmarks:
            cv2.circle(img, (int(x), int(y)), 10, (255, 0, 0), -1)

        plt.imshow(img)
        plt.show()



import importlib
import os
import numpy as np
import cv2
import torch.nn.parallel
import torch.utils.data
import torchvision.transforms as transforms
from PIL import Image

import torchvision.models as models
from visage.modules.FaceBoxesV2.faceboxes_detector import FaceBoxesDetector
from visage.modules.PIPNet.lib.functions import get_meanface, forward_pip
from visage.modules.PIPNet.lib.mobilenetv3 import mobilenetv3_large
from visage.modules.PIPNet.lib.networks import Pip_resnet18, Pip_resnet50, Pip_resnet101, Pip_mbnetv2, Pip_mbnetv3


def demo_image(images, net, preprocess, cfg, input_size, net_stride, num_nb, use_gpu, device):
    snapshot_dir = os.path.join(os.sep, os.path.join(*__file__.split(os.path.sep)[:-2]))
    detector = FaceBoxesDetector('FaceBoxes', os.path.join(snapshot_dir, 'FaceBoxesV2/weights/FaceBoxesV2.pth'),
                                 use_gpu, device)
    my_thresh = 0.6
    det_box_scale = 1.2
    meanface_indices, reverse_index1, reverse_index2, max_len = get_meanface(
        os.path.join(snapshot_dir, 'data', cfg.data_name, 'meanface.txt'), cfg.num_nb)

    net.eval()

    print('#########################################################')
    landmarks = []
    annotated_images = []
    bboxes = []
    confidences = []
    for image in images:
        image = image.copy()
        image_height, image_width, _ = image.shape
        detections, _ = detector.detect(image, my_thresh, 1)
        pred_export = None
        dets_filtered = [det for det in detections if det[0] == 'face']
        dets_filtered.sort(key=lambda x: -1 * x[1])
        detections = dets_filtered
        print(detections)
        for i in range(min(1, len(detections))):
            det_xmin = detections[i][2]
            det_ymin = detections[i][3]
            det_width = detections[i][4]
            det_height = detections[i][5]
            det_xmax = det_xmin + det_width - 1
            det_ymax = det_ymin + det_height - 1

            det_xmin -= int(det_width * (det_box_scale - 1) / 2)
            # remove a part of top area for alignment, see paper for details
            det_ymin += int(det_height * (det_box_scale - 1) / 2)
            det_xmax += int(det_width * (det_box_scale - 1) / 2)
            det_ymax += int(det_height * (det_box_scale - 1) / 2)
            det_xmin = max(det_xmin, 0)
            det_ymin = max(det_ymin, 0)
            det_xmax = min(det_xmax, image_width - 1)
            det_ymax = min(det_ymax, image_height - 1)
            det_width = det_xmax - det_xmin + 1
            det_height = det_ymax - det_ymin + 1
            cv2.rectangle(image, (det_xmin, det_ymin), (det_xmax, det_ymax), (0, 0, 255), 2)
            det_crop = image[det_ymin:det_ymax, det_xmin:det_xmax, :]
            det_crop = cv2.resize(det_crop, (input_size, input_size))
            inputs = Image.fromarray(det_crop[:, :, ::-1].astype('uint8'), 'RGB')
            inputs = preprocess(inputs).unsqueeze(0)
            inputs = inputs.to(device)
            lms_pred_x, lms_pred_y, lms_pred_nb_x, lms_pred_nb_y, outputs_cls, max_cls = forward_pip(net, inputs,
                                                                                                     preprocess,
                                                                                                     input_size,
                                                                                                     net_stride, num_nb)
            lms_pred = torch.cat((lms_pred_x, lms_pred_y), dim=1).flatten()
            tmp_nb_x = lms_pred_nb_x[reverse_index1, reverse_index2].view(cfg.num_lms, max_len)
            tmp_nb_y = lms_pred_nb_y[reverse_index1, reverse_index2].view(cfg.num_lms, max_len)
            tmp_x = torch.mean(torch.cat((lms_pred_x, tmp_nb_x), dim=1), dim=1).view(-1, 1)
            tmp_y = torch.mean(torch.cat((lms_pred_y, tmp_nb_y), dim=1), dim=1).view(-1, 1)
            lms_pred_merge = torch.cat((tmp_x, tmp_y), dim=1).flatten()
            lms_pred = lms_pred.cpu().numpy()
            lms_pred_merge = lms_pred_merge.cpu().numpy()
            pred_export = np.zeros([cfg.num_lms, 2])
            for ii in range(cfg.num_lms):
                x_pred = lms_pred_merge[ii * 2] * det_width
                y_pred = lms_pred_merge[ii * 2 + 1] * det_height
                pred_export[ii, 0] = x_pred + det_xmin
                pred_export[ii, 1] = y_pred + det_ymin
                cv2.circle(image, (int(x_pred) + det_xmin, int(y_pred) + det_ymin), 1, (0, 0, 255), 2)
                if ii == 76:
                    cv2.circle(image, (int(x_pred) + det_xmin, int(y_pred) + det_ymin), 1, (255, 0, 0), 2)
            landmarks.append(pred_export)
            annotated_images.append(image)
            bboxes.append(detections[i][2:])
            confidences.append(detections[i][1])

    return landmarks, annotated_images, bboxes, confidences


def face_detection(images, use_gpu, device):
    snapshot_dir = os.path.join(os.sep, os.path.join(*__file__.split(os.path.sep)[:-2]))
    detector = FaceBoxesDetector('FaceBoxes', os.path.join(snapshot_dir, 'FaceBoxesV2/weights/FaceBoxesV2.pth'),
                                 use_gpu, device)
    my_thresh = 0.6

    print('#########################################################')
    bboxes = []
    confidences = []
    for image in images:
        image = image.copy()
        image_height, image_width, _ = image.shape
        detections, _ = detector.detect(image, my_thresh, 1)
        dets_filtered = [det for det in detections if det[0] == 'face']
        dets_filtered.sort(key=lambda x: -1 * x[1])
        detections = dets_filtered
        print(detections)
        for i in range(min(1, len(detections))):
            bboxes.append(detections[i][2:])
            cur_det = detections[i][1]

            confidences.append(cur_det)
        if len(detections) < 1:
            bboxes.append([-1, -1, -1, -1])
            confidences.append(-1)

    return bboxes, confidences


def landmark_detection(images, bboxes, exp_path, device):
    experiment_name = exp_path.split('/')[-1][:-3]
    data_name = exp_path.split('/')[-2]
    config_path = '.experiments.{}.{}'.format(data_name, experiment_name)

    my_config = importlib.import_module(config_path, package='famudy.preprocessing.PIPNet')
    Config = getattr(my_config, 'Config')
    cfg = Config()
    cfg.experiment_name = experiment_name
    cfg.data_name = data_name
    snapshot_dir = os.path.join(os.sep, os.path.join(*__file__.split(os.path.sep)[:-2] + ['snapshots']))
    save_dir = os.path.join(snapshot_dir, cfg.data_name, cfg.experiment_name)

    if cfg.backbone == 'resnet18':
        resnet18 = models.resnet18(pretrained=cfg.pretrained)
        net = Pip_resnet18(resnet18, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                           net_stride=cfg.net_stride)
    elif cfg.backbone == 'resnet50':
        resnet50 = models.resnet50(pretrained=cfg.pretrained)
        net = Pip_resnet50(resnet50, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                           net_stride=cfg.net_stride)
    elif cfg.backbone == 'resnet101':
        resnet101 = models.resnet101(pretrained=cfg.pretrained)
        net = Pip_resnet101(resnet101, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                            net_stride=cfg.net_stride)
    elif cfg.backbone == 'mobilenet_v2':
        mbnet = models.mobilenet_v2(pretrained=cfg.pretrained)
        net = Pip_mbnetv2(mbnet, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size, net_stride=cfg.net_stride)
    elif cfg.backbone == 'mobilenet_v3':
        mbnet = mobilenetv3_large()
        if cfg.pretrained:
            mbnet.load_state_dict(torch.load('lib/mobilenetv3-large-1cd25616.pth'))
        net = Pip_mbnetv3(mbnet, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size, net_stride=cfg.net_stride)
    else:
        print('No such backbone!')
        exit(0)

    net = net.to(device)

    weight_file = os.path.join(save_dir, 'epoch%d.pth' % (cfg.num_epochs - 1))
    state_dict = torch.load(weight_file, map_location=device)
    net.load_state_dict(state_dict)

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
    preprocess = transforms.Compose(
        [transforms.Resize((cfg.input_size, cfg.input_size)), transforms.ToTensor(), normalize])

    snapshot_dir = os.path.join(os.sep, os.path.join(*__file__.split(os.path.sep)[:-2]))
    det_box_scale = 1.2
    meanface_indices, reverse_index1, reverse_index2, max_len = get_meanface(
        os.path.join(snapshot_dir, 'data', cfg.data_name, 'meanface.txt'), cfg.num_nb)

    net.eval()

    print('#########################################################')
    landmarks = []
    annotated_images = []
    for im_idx, image in enumerate(images):
        if bboxes[im_idx][-1] == -1:
            landmarks.append(None)
            continue
        image = image.copy()
        image_height, image_width, _ = image.shape
        det_xmin = int(bboxes[im_idx][0])
        det_ymin = int(bboxes[im_idx][1])
        det_width = int(bboxes[im_idx][2])
        det_height = int(bboxes[im_idx][3])
        det_xmax = det_xmin + det_width - 1
        det_ymax = det_ymin + det_height - 1

        det_xmin -= int(det_width * (det_box_scale - 1) / 2)
        # remove a part of top area for alignment, see paper for details
        det_ymin += int(det_height * (det_box_scale - 1) / 2)
        det_xmax += int(det_width * (det_box_scale - 1) / 2)
        det_ymax += int(det_height * (det_box_scale - 1) / 2)
        det_xmin = max(det_xmin, 0)
        det_ymin = max(det_ymin, 0)
        det_xmax = min(det_xmax, image_width - 1)
        det_ymax = min(det_ymax, image_height - 1)
        det_width = det_xmax - det_xmin + 1
        det_height = det_ymax - det_ymin + 1
        # cv2.rectangle(image, (det_xmin, det_ymin), (det_xmax, det_ymax), (0, 0, 255), 2)
        det_crop = image[det_ymin:det_ymax, det_xmin:det_xmax, :]
        # plt.imshow(det_crop)
        # plt.show()
        det_crop = cv2.resize(det_crop, (cfg.input_size, cfg.input_size))
        inputs = Image.fromarray(det_crop[:, :, ::-1].astype('uint8'), 'RGB')
        inputs = preprocess(inputs).unsqueeze(0)
        inputs = inputs.to(device)
        lms_pred_x, lms_pred_y, lms_pred_nb_x, lms_pred_nb_y, outputs_cls, max_cls = forward_pip(net, inputs,
                                                                                                 preprocess,
                                                                                                 cfg.input_size,
                                                                                                 cfg.net_stride,
                                                                                                 cfg.num_nb)
        tmp_nb_x = lms_pred_nb_x[reverse_index1, reverse_index2].view(cfg.num_lms, max_len)
        tmp_nb_y = lms_pred_nb_y[reverse_index1, reverse_index2].view(cfg.num_lms, max_len)
        tmp_x = torch.mean(torch.cat((lms_pred_x, tmp_nb_x), dim=1), dim=1).view(-1, 1)
        tmp_y = torch.mean(torch.cat((lms_pred_y, tmp_nb_y), dim=1), dim=1).view(-1, 1)
        lms_pred_merge = torch.cat((tmp_x, tmp_y), dim=1).flatten()
        lms_pred_merge = lms_pred_merge.cpu().numpy()
        pred_export = np.zeros([cfg.num_lms, 2])
        for ii in range(cfg.num_lms):
            x_pred = lms_pred_merge[ii * 2] * det_width
            y_pred = lms_pred_merge[ii * 2 + 1] * det_height
            pred_export[ii, 0] = x_pred + det_xmin
            pred_export[ii, 1] = y_pred + det_ymin
        landmarks.append(pred_export)

    hit = None
    for i in range(len(landmarks)):
        if landmarks[i] is None:
            if hit is not None:
                landmarks[i] = -np.ones_like(hit)
                continue
            j = i + 1
            while landmarks[j] is None:
                j += 1
            hit = landmarks[j]
            landmarks[i] = -np.ones_like(hit)

    return np.stack(landmarks, axis=0)


def run(exp_path, images, device):
    experiment_name = exp_path.split('/')[-1][:-3]
    data_name = exp_path.split('/')[-2]
    config_path = '.experiments.{}.{}'.format(data_name, experiment_name)

    my_config = importlib.import_module(config_path, package='famudy.preprocessing.PIPNet')
    Config = getattr(my_config, 'Config')
    cfg = Config()
    cfg.experiment_name = experiment_name
    cfg.data_name = data_name
    snapshot_dir = os.path.join(os.sep, os.path.join(*__file__.split(os.path.sep)[:-2] + ['snapshots']))
    save_dir = os.path.join(snapshot_dir, cfg.data_name, cfg.experiment_name)

    if cfg.backbone == 'resnet18':
        resnet18 = models.resnet18(pretrained=cfg.pretrained)
        net = Pip_resnet18(resnet18, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                           net_stride=cfg.net_stride)
    elif cfg.backbone == 'resnet50':
        resnet50 = models.resnet50(pretrained=cfg.pretrained)
        net = Pip_resnet50(resnet50, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                           net_stride=cfg.net_stride)
    elif cfg.backbone == 'resnet101':
        resnet101 = models.resnet101(pretrained=cfg.pretrained)
        net = Pip_resnet101(resnet101, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size,
                            net_stride=cfg.net_stride)
    elif cfg.backbone == 'mobilenet_v2':
        mbnet = models.mobilenet_v2(pretrained=cfg.pretrained)
        net = Pip_mbnetv2(mbnet, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size, net_stride=cfg.net_stride)
    elif cfg.backbone == 'mobilenet_v3':
        mbnet = mobilenetv3_large()
        if cfg.pretrained:
            mbnet.load_state_dict(torch.load('lib/mobilenetv3-large-1cd25616.pth'))
        net = Pip_mbnetv3(mbnet, cfg.num_nb, num_lms=cfg.num_lms, input_size=cfg.input_size, net_stride=cfg.net_stride)
    else:
        print('No such backbone!')
        exit(0)

    net = net.to(device)

    weight_file = os.path.join(save_dir, 'epoch%d.pth' % (cfg.num_epochs - 1))
    state_dict = torch.load(weight_file, map_location=device)
    net.load_state_dict(state_dict)

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
    preprocess = transforms.Compose(
        [transforms.Resize((cfg.input_size, cfg.input_size)), transforms.ToTensor(), normalize])

    return demo_image(images, net, preprocess, cfg, cfg.input_size, cfg.net_stride, cfg.num_nb, cfg.use_gpu, device)


if __name__ == '__main__':
    os.makedirs('/mnt/hdd/test_pip', exist_ok=True)
    run('/mnt/hdd/test_pip', '~/famudy/pips/demo_images3/')



import cv2
import numpy as np
import torch


def apply_colormap(segmentation_mask: torch.tensor, stride: int = 1) -> np.array:
    # Colors for all 20 parts
    part_colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0],
                   [255, 0, 85], [255, 0, 170],
                   [0, 255, 0], [85, 255, 0], [170, 255, 0],
                   [0, 255, 85], [0, 255, 170],
                   [0, 0, 255], [85, 0, 255], [170, 0, 255],
                   [0, 85, 255], [0, 170, 255],
                   [255, 255, 0], [255, 255, 85], [255, 255, 170],
                   [255, 0, 255], [255, 85, 255], [255, 170, 255],
                   [0, 255, 255], [85, 255, 255], [170, 255, 255]]

    if isinstance(segmentation_mask, torch.Tensor):
        segmentation_mask = segmentation_mask.numpy()

    vis_parsing_anno = segmentation_mask.copy().astype(np.uint8)
    vis_parsing_anno = cv2.resize(vis_parsing_anno, None, fx=stride, fy=stride, interpolation=cv2.INTER_NEAREST)
    vis_parsing_anno_color = np.zeros((vis_parsing_anno.shape[0], vis_parsing_anno.shape[1], 3)) + 255

    num_of_class = np.max(vis_parsing_anno)

    for pi in range(1, num_of_class + 1):
        index = np.where(vis_parsing_anno == pi)
        vis_parsing_anno_color[index[0], index[1], :] = part_colors[pi]

    vis_parsing_anno_color = vis_parsing_anno_color.astype(np.uint8)

    return vis_parsing_anno_color



from unittest import TestCase

import numpy as np

from visage.evaluator.paired_face_image_evaluator import PairedFaceImageEvaluator
from visage.evaluator.paired_image_evaluator import PairedImageEvaluator


class EvaluationTest(TestCase):

    def test_paired_image_evaluator(self):
        paired_image_evaluator = PairedImageEvaluator()

        predictions = np.zeros((4, 512, 512, 3))
        targets = np.ones((4, 512, 512, 3))

        paired_image_metrics = paired_image_evaluator.evaluate(predictions, targets)

        print(paired_image_metrics)

    def test_paired_face_image_evaluator(self):
        paired_face_image_evaluator = PairedFaceImageEvaluator()

        predictions = np.zeros((4, 512, 512, 3))
        targets = np.ones((4, 512, 512, 3))

        paired_face_image_metrics = paired_face_image_evaluator.evaluate(predictions, targets)

        print(paired_face_image_metrics)



from dataclasses import dataclass


@dataclass
class PIPNetConfig:
    backbone: str
    pretrained: bool
    num_nb: int
    num_lms: int  # Number of landmarks
    input_size: int
    net_stride: int
    checkpoint_path: str  # Path inside $HOME/.cache/torch/face-parser/PIPNet where checkpoint is located
    meanface_path: str  # Path inside face_parser.PIPNet.meanfaces where meanface specifications are stored


PIPNet_WFLW_r18_config = PIPNetConfig(
    backbone='resnet18',
    pretrained=True,
    num_nb=10,
    num_lms=98,
    input_size=256,
    net_stride=32,
    checkpoint_path='WFLW/pip_32_16_60_r18_l2_l1_10_1_nb10/epoch59.pth',
    meanface_path='WFLW/meanface.txt'
)



from dataclasses import dataclass, asdict
from typing import Optional, List

import numpy as np
import torch

from visage.evaluator.csim_evaluator import CSIMEvaluator
from visage.evaluator.keypoint_evaluator import KeypointEvaluator
from visage.evaluator.paired_image_evaluator import PairedImageMetrics, PairedImageEvaluator


@dataclass
class PairedFaceImageMetrics(PairedImageMetrics):
    akd: Optional[float] = None
    akd_face: Optional[float] = None
    csim: Optional[float] = None
    aed: Optional[float] = None
    apd: Optional[float] = None

    def __add__(self, other: 'PairedFaceImageMetrics') -> 'PairedFaceImageMetrics':
        paired_image_metrics = super().__add__(other)
        return PairedFaceImageMetrics(
            **asdict(paired_image_metrics),
            akd=self.akd + other.akd if self.akd is not None and other.akd is not None else None,
            akd_face=self.akd_face + other.akd_face if self.akd_face is not None and other.akd_face is not None else None,
            csim=self.csim + other.csim if self.csim is not None and other.csim is not None else None,
            aed=self.aed + other.aed if self.aed is not None and other.aed is not None else None,
            apd=self.apd + other.apd if self.apd is not None and other.apd is not None else None,
        )

    def __radd__(self, other) -> 'PairedFaceImageMetrics':
        if other == 0:
            return self

        return self + other

    def __truediv__(self, scalar: float) -> 'PairedFaceImageMetrics':
        paired_image_metrics = super().__truediv__(scalar)
        return PairedFaceImageMetrics(
            **asdict(paired_image_metrics),
            akd=self.akd / scalar if self.akd is not None else None,
            akd_face=self.akd_face / scalar if self.akd_face is not None else None,
            csim=self.csim / scalar if self.csim is not None else None,
            aed=self.aed / scalar if self.aed is not None else None,
            apd=self.apd / scalar if self.apd is not None else None,
        )


class PairedFaceImageEvaluator(PairedImageEvaluator):

    def __init__(self, exclude_lpips: bool = False):
        super().__init__(exclude_lpips=exclude_lpips)
        self._keypoint_evaluator = KeypointEvaluator()
        self._csim_evaluator = CSIMEvaluator()

    def evaluate(self, predictions: List[np.ndarray], targets: List[np.ndarray]) -> PairedFaceImageMetrics:
        return super().evaluate(predictions, targets)

    def _evaluate(self, predictions_torch: torch.Tensor, targets_torch: torch.Tensor) -> PairedFaceImageMetrics:
        paired_image_metrics = super()._evaluate(predictions_torch, targets_torch)

        akd, akd_face = self._keypoint_evaluator(predictions_torch, targets_torch)
        csim = self._csim_evaluator(predictions_torch, targets_torch)

        return PairedFaceImageMetrics(**asdict(paired_image_metrics), akd=akd.item(), akd_face=akd_face.item(), csim=csim.item())



import torch
from matplotlib import pyplot as plt

from base import VisionTestCase
from visage.bisenet import BiSeNetFaceParser
from visage.visualize import apply_colormap


class BiSeNetTest(VisionTestCase):

    def test_parse(self):
        face_parser = BiSeNetFaceParser()

        B = 5
        H = 71
        W = 83

        images = torch.randn((B, H, W, 3))
        segmentation_masks = face_parser.parse(images)

        self.assertEqual(len(segmentation_masks.shape), 3)
        self.assertEqual(segmentation_masks.shape[0], B)
        self.assertEqual(segmentation_masks.shape[1], H)
        self.assertEqual(segmentation_masks.shape[2], W)

    def test_real_image(self):
        image = self._get_obama_image(keep_aspect_ratio=True, resize=512)

        with torch.no_grad():
            face_parser = BiSeNetFaceParser()

            segmentation_mask = face_parser.parse(image)
            colormap = apply_colormap(segmentation_mask)

            self.assertEqual(colormap.shape[0], image.shape[1])
            self.assertEqual(colormap.shape[1], image.shape[2])
            self.assertEqual(colormap.shape[2], image.shape[0])

            plt.imshow(0.5 * colormap / 255 + 0.5 * (image.permute(1, 2, 0).numpy() + 1) / 2)
            plt.show()



from typing import Tuple

import numpy as np
import torch
from dreifus.image import Img
from visage.bounding_boxes.face_boxes_v2 import FaceBoxesV2
from visage.landmark_detection.pipnet import PIPNet


class KeypointEvaluator:

    def __init__(self):
        self._detector = FaceBoxesV2()
        self._pip_net = PIPNet()

    def __call__(self, predictions: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        keypoint_distances = []
        keypoint_distances_face = []
        for prediction, target in zip(predictions, targets):
            prediction = Img.from_torch(prediction).to_numpy().img
            target = Img.from_torch(target).to_numpy().img
            detected_bboxes_prediction = self._detector(prediction)
            detected_bboxes_target = self._detector(target)
            if len(detected_bboxes_prediction) > 0 and len(detected_bboxes_target) > 0:
                landmarks_prediction = self._pip_net.forward(prediction, detected_bboxes_prediction[0])
                landmarks_target = self._pip_net.forward(target, detected_bboxes_target[0])

                keypoint_distance = np.linalg.norm(landmarks_prediction - landmarks_target, axis=-1).mean()
                keypoint_distance_face = np.linalg.norm(landmarks_prediction[33:] - landmarks_target[33:], axis=-1).mean()
                keypoint_distances.append(keypoint_distance)
                keypoint_distances_face.append(keypoint_distance_face)

        if len(keypoint_distances) > 0:
            average_keypoint_distance = torch.tensor(np.stack(keypoint_distances).mean())
            average_keypoint_distance_face = torch.tensor(np.stack(keypoint_distances_face).mean())
        else:
            average_keypoint_distance = torch.tensor(-1)
            average_keypoint_distance_face = torch.tensor(-1)

        return average_keypoint_distance, average_keypoint_distance_face



import torch
from elias.util.batch import batchify_sliced
from torch import nn
from torchmetrics.image import MultiScaleStructuralSimilarityIndexMeasure, StructuralSimilarityIndexMeasure


class SSIMEvaluator(nn.Module):

    def __init__(self, batch_size: int = 16):
        super(SSIMEvaluator, self).__init__()
        self._ssim_evaluator = StructuralSimilarityIndexMeasure(data_range=1.0)
        self._batch_size = batch_size

    def __call__(self, predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        ssims = []
        for predictions_batch, targets_batch in zip(batchify_sliced(predictions, self._batch_size), batchify_sliced(targets, self._batch_size)):
            ssims.append(self._ssim_evaluator(predictions_batch, targets_batch))

        ssim = torch.stack(ssims).mean()
        return ssim


class MultiScaleSSIMEvaluator(nn.Module):

    def __init__(self, batch_size: int = 16):
        super(MultiScaleSSIMEvaluator, self).__init__()
        self._batch_size = batch_size
        self._multi_scale_ssim_evaluator = MultiScaleStructuralSimilarityIndexMeasure(data_range=1.0)

    def __call__(self, predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        multi_scale_ssims = []
        for predictions_batch, targets_batch in zip(batchify_sliced(predictions, self._batch_size), batchify_sliced(targets, self._batch_size)):
            multi_scale_ssims.append(self._multi_scale_ssim_evaluator(predictions_batch, targets_batch))

        multi_scale_ssim = torch.stack(multi_scale_ssims).mean()
        return multi_scale_ssim



import torch
from elias.util.batch import batchify_sliced
from torch import nn
from torchmetrics import PeakSignalNoiseRatio


class PSNREvaluator(nn.Module):

    def __init__(self, batch_size: int = 16):
        super(PSNREvaluator, self).__init__()
        self._psnr_evaluator = PeakSignalNoiseRatio(data_range=1.0)
        self._batch_size = batch_size

    def __call__(self, predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        psnrs = []
        for predictions_batch, targets_batch in zip(batchify_sliced(predictions, self._batch_size), batchify_sliced(targets, self._batch_size)):
            psnrs.append(self._psnr_evaluator(predictions_batch, targets_batch))

        psnr = torch.stack(psnrs).mean()
        return psnr



import torch.nn as nn
import torch.nn.functional as F


# net_stride output_size
# 128        2x2
# 64         4x4
# 32         8x8
# pip regression, resnet101
class Pip_resnet101(nn.Module):
    def __init__(self, resnet, num_nb, num_lms=68, input_size=256, net_stride=32):
        super(Pip_resnet101, self).__init__()
        self.num_nb = num_nb
        self.num_lms = num_lms
        self.input_size = input_size
        self.net_stride = net_stride
        self.conv1 = resnet.conv1
        self.bn1 = resnet.bn1
        self.maxpool = resnet.maxpool
        self.sigmoid = nn.Sigmoid()
        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4
        if self.net_stride == 128:
            self.layer5 = nn.Conv2d(2048, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            self.layer6 = nn.Conv2d(512, 512, kernel_size=3, stride=2, padding=1)
            self.bn6 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)

            nn.init.normal_(self.layer6.weight, std=0.001)
            if self.layer6.bias is not None:
                nn.init.constant_(self.layer6.bias, 0)
            nn.init.constant_(self.bn6.weight, 1)
            nn.init.constant_(self.bn6.bias, 0)
        elif self.net_stride == 64:
            self.layer5 = nn.Conv2d(2048, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)
        elif self.net_stride == 32:
            pass
        else:
            print('No such net_stride!')
            exit(0)

        self.cls_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.x_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.y_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_x_layer = nn.Conv2d(2048, num_nb * num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_y_layer = nn.Conv2d(2048, num_nb * num_lms, kernel_size=1, stride=1, padding=0)

        nn.init.normal_(self.cls_layer.weight, std=0.001)
        if self.cls_layer.bias is not None:
            nn.init.constant_(self.cls_layer.bias, 0)

        nn.init.normal_(self.x_layer.weight, std=0.001)
        if self.x_layer.bias is not None:
            nn.init.constant_(self.x_layer.bias, 0)

        nn.init.normal_(self.y_layer.weight, std=0.001)
        if self.y_layer.bias is not None:
            nn.init.constant_(self.y_layer.bias, 0)

        nn.init.normal_(self.nb_x_layer.weight, std=0.001)
        if self.nb_x_layer.bias is not None:
            nn.init.constant_(self.nb_x_layer.bias, 0)

        nn.init.normal_(self.nb_y_layer.weight, std=0.001)
        if self.nb_y_layer.bias is not None:
            nn.init.constant_(self.nb_y_layer.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self.net_stride == 128:
            x = F.relu(self.bn5(self.layer5(x)))
            x = F.relu(self.bn6(self.layer6(x)))
        elif self.net_stride == 64:
            x = F.relu(self.bn5(self.layer5(x)))
        else:
            pass
        x1 = self.cls_layer(x)
        x2 = self.x_layer(x)
        x3 = self.y_layer(x)
        x4 = self.nb_x_layer(x)
        x5 = self.nb_y_layer(x)
        return x1, x2, x3, x4, x5


# net_stride output_size
# 128        2x2
# 64         4x4
# 32         8x8
# pip regression, resnet50
class Pip_resnet50(nn.Module):
    def __init__(self, resnet, num_nb, num_lms=68, input_size=256, net_stride=32):
        super(Pip_resnet50, self).__init__()
        self.num_nb = num_nb
        self.num_lms = num_lms
        self.input_size = input_size
        self.net_stride = net_stride
        self.conv1 = resnet.conv1
        self.bn1 = resnet.bn1
        self.maxpool = resnet.maxpool
        self.sigmoid = nn.Sigmoid()
        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4
        if self.net_stride == 128:
            self.layer5 = nn.Conv2d(2048, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            self.layer6 = nn.Conv2d(512, 512, kernel_size=3, stride=2, padding=1)
            self.bn6 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)

            nn.init.normal_(self.layer6.weight, std=0.001)
            if self.layer6.bias is not None:
                nn.init.constant_(self.layer6.bias, 0)
            nn.init.constant_(self.bn6.weight, 1)
            nn.init.constant_(self.bn6.bias, 0)
        elif self.net_stride == 64:
            self.layer5 = nn.Conv2d(2048, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)
        elif self.net_stride == 32:
            pass
        else:
            print('No such net_stride!')
            exit(0)

        self.cls_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.x_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.y_layer = nn.Conv2d(2048, num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_x_layer = nn.Conv2d(2048, num_nb * num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_y_layer = nn.Conv2d(2048, num_nb * num_lms, kernel_size=1, stride=1, padding=0)

        nn.init.normal_(self.cls_layer.weight, std=0.001)
        if self.cls_layer.bias is not None:
            nn.init.constant_(self.cls_layer.bias, 0)

        nn.init.normal_(self.x_layer.weight, std=0.001)
        if self.x_layer.bias is not None:
            nn.init.constant_(self.x_layer.bias, 0)

        nn.init.normal_(self.y_layer.weight, std=0.001)
        if self.y_layer.bias is not None:
            nn.init.constant_(self.y_layer.bias, 0)

        nn.init.normal_(self.nb_x_layer.weight, std=0.001)
        if self.nb_x_layer.bias is not None:
            nn.init.constant_(self.nb_x_layer.bias, 0)

        nn.init.normal_(self.nb_y_layer.weight, std=0.001)
        if self.nb_y_layer.bias is not None:
            nn.init.constant_(self.nb_y_layer.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self.net_stride == 128:
            x = F.relu(self.bn5(self.layer5(x)))
            x = F.relu(self.bn6(self.layer6(x)))
        elif self.net_stride == 64:
            x = F.relu(self.bn5(self.layer5(x)))
        else:
            pass
        x1 = self.cls_layer(x)
        x2 = self.x_layer(x)
        x3 = self.y_layer(x)
        x4 = self.nb_x_layer(x)
        x5 = self.nb_y_layer(x)
        return x1, x2, x3, x4, x5


# net_stride output_size
# 128        2x2
# 64         4x4
# 32         8x8
# pip regression, resnet18
class Pip_resnet18(nn.Module):
    def __init__(self, resnet, num_nb, num_lms=68, input_size=256, net_stride=32):
        super(Pip_resnet18, self).__init__()
        self.num_nb = num_nb
        self.num_lms = num_lms
        self.input_size = input_size
        self.net_stride = net_stride
        self.conv1 = resnet.conv1
        self.bn1 = resnet.bn1
        self.maxpool = resnet.maxpool
        self.sigmoid = nn.Sigmoid()
        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4
        if self.net_stride == 128:
            self.layer5 = nn.Conv2d(512, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            self.layer6 = nn.Conv2d(512, 512, kernel_size=3, stride=2, padding=1)
            self.bn6 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)

            nn.init.normal_(self.layer6.weight, std=0.001)
            if self.layer6.bias is not None:
                nn.init.constant_(self.layer6.bias, 0)
            nn.init.constant_(self.bn6.weight, 1)
            nn.init.constant_(self.bn6.bias, 0)
        elif self.net_stride == 64:
            self.layer5 = nn.Conv2d(512, 512, kernel_size=3, stride=2, padding=1)
            self.bn5 = nn.BatchNorm2d(512)
            # init
            nn.init.normal_(self.layer5.weight, std=0.001)
            if self.layer5.bias is not None:
                nn.init.constant_(self.layer5.bias, 0)
            nn.init.constant_(self.bn5.weight, 1)
            nn.init.constant_(self.bn5.bias, 0)
        elif self.net_stride == 32:
            pass
        elif self.net_stride == 16:
            self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=4, stride=2, padding=1, bias=False)
            self.bn_deconv1 = nn.BatchNorm2d(512)
            nn.init.normal_(self.deconv1.weight, std=0.001)
            if self.deconv1.bias is not None:
                nn.init.constant_(self.deconv1.bias, 0)
            nn.init.constant_(self.bn_deconv1.weight, 1)
            nn.init.constant_(self.bn_deconv1.bias, 0)
        else:
            print('No such net_stride!')
            exit(0)

        self.cls_layer = nn.Conv2d(512, num_lms, kernel_size=1, stride=1, padding=0)
        self.x_layer = nn.Conv2d(512, num_lms, kernel_size=1, stride=1, padding=0)
        self.y_layer = nn.Conv2d(512, num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_x_layer = nn.Conv2d(512, num_nb * num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_y_layer = nn.Conv2d(512, num_nb * num_lms, kernel_size=1, stride=1, padding=0)

        nn.init.normal_(self.cls_layer.weight, std=0.001)
        if self.cls_layer.bias is not None:
            nn.init.constant_(self.cls_layer.bias, 0)

        nn.init.normal_(self.x_layer.weight, std=0.001)
        if self.x_layer.bias is not None:
            nn.init.constant_(self.x_layer.bias, 0)

        nn.init.normal_(self.y_layer.weight, std=0.001)
        if self.y_layer.bias is not None:
            nn.init.constant_(self.y_layer.bias, 0)

        nn.init.normal_(self.nb_x_layer.weight, std=0.001)
        if self.nb_x_layer.bias is not None:
            nn.init.constant_(self.nb_x_layer.bias, 0)

        nn.init.normal_(self.nb_y_layer.weight, std=0.001)
        if self.nb_y_layer.bias is not None:
            nn.init.constant_(self.nb_y_layer.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self.net_stride == 128:
            x = F.relu(self.bn5(self.layer5(x)))
            x = F.relu(self.bn6(self.layer6(x)))
        elif self.net_stride == 64:
            x = F.relu(self.bn5(self.layer5(x)))
        elif self.net_stride == 16:
            x = F.relu(self.bn_deconv1(self.deconv1(x)))
        else:
            pass
        x1 = self.cls_layer(x)
        x2 = self.x_layer(x)
        x3 = self.y_layer(x)
        x4 = self.nb_x_layer(x)
        x5 = self.nb_y_layer(x)
        return x1, x2, x3, x4, x5


class Pip_mbnetv2(nn.Module):
    def __init__(self, mbnet, num_nb, num_lms=68, input_size=256, net_stride=32):
        super(Pip_mbnetv2, self).__init__()
        self.num_nb = num_nb
        self.num_lms = num_lms
        self.input_size = input_size
        self.net_stride = net_stride
        self.features = mbnet.features
        self.sigmoid = nn.Sigmoid()

        self.cls_layer = nn.Conv2d(1280, num_lms, kernel_size=1, stride=1, padding=0)
        self.x_layer = nn.Conv2d(1280, num_lms, kernel_size=1, stride=1, padding=0)
        self.y_layer = nn.Conv2d(1280, num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_x_layer = nn.Conv2d(1280, num_nb * num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_y_layer = nn.Conv2d(1280, num_nb * num_lms, kernel_size=1, stride=1, padding=0)

        nn.init.normal_(self.cls_layer.weight, std=0.001)
        if self.cls_layer.bias is not None:
            nn.init.constant_(self.cls_layer.bias, 0)

        nn.init.normal_(self.x_layer.weight, std=0.001)
        if self.x_layer.bias is not None:
            nn.init.constant_(self.x_layer.bias, 0)

        nn.init.normal_(self.y_layer.weight, std=0.001)
        if self.y_layer.bias is not None:
            nn.init.constant_(self.y_layer.bias, 0)

        nn.init.normal_(self.nb_x_layer.weight, std=0.001)
        if self.nb_x_layer.bias is not None:
            nn.init.constant_(self.nb_x_layer.bias, 0)

        nn.init.normal_(self.nb_y_layer.weight, std=0.001)
        if self.nb_y_layer.bias is not None:
            nn.init.constant_(self.nb_y_layer.bias, 0)

    def forward(self, x):
        x = self.features(x)
        x1 = self.cls_layer(x)
        x2 = self.x_layer(x)
        x3 = self.y_layer(x)
        x4 = self.nb_x_layer(x)
        x5 = self.nb_y_layer(x)
        return x1, x2, x3, x4, x5


class Pip_mbnetv3(nn.Module):
    def __init__(self, mbnet, num_nb, num_lms=68, input_size=256, net_stride=32):
        super(Pip_mbnetv3, self).__init__()
        self.num_nb = num_nb
        self.num_lms = num_lms
        self.input_size = input_size
        self.net_stride = net_stride
        self.features = mbnet.features
        self.conv = mbnet.conv
        self.sigmoid = nn.Sigmoid()

        self.cls_layer = nn.Conv2d(960, num_lms, kernel_size=1, stride=1, padding=0)
        self.x_layer = nn.Conv2d(960, num_lms, kernel_size=1, stride=1, padding=0)
        self.y_layer = nn.Conv2d(960, num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_x_layer = nn.Conv2d(960, num_nb * num_lms, kernel_size=1, stride=1, padding=0)
        self.nb_y_layer = nn.Conv2d(960, num_nb * num_lms, kernel_size=1, stride=1, padding=0)

        nn.init.normal_(self.cls_layer.weight, std=0.001)
        if self.cls_layer.bias is not None:
            nn.init.constant_(self.cls_layer.bias, 0)

        nn.init.normal_(self.x_layer.weight, std=0.001)
        if self.x_layer.bias is not None:
            nn.init.constant_(self.x_layer.bias, 0)

        nn.init.normal_(self.y_layer.weight, std=0.001)
        if self.y_layer.bias is not None:
            nn.init.constant_(self.y_layer.bias, 0)

        nn.init.normal_(self.nb_x_layer.weight, std=0.001)
        if self.nb_x_layer.bias is not None:
            nn.init.constant_(self.nb_x_layer.bias, 0)

        nn.init.normal_(self.nb_y_layer.weight, std=0.001)
        if self.nb_y_layer.bias is not None:
            nn.init.constant_(self.nb_y_layer.bias, 0)

    def forward(self, x):
        x = self.features(x)
        x = self.conv(x)
        x1 = self.cls_layer(x)
        x2 = self.x_layer(x)
        x3 = self.y_layer(x)
        x4 = self.nb_x_layer(x)
        x5 = self.nb_y_layer(x)
        return x1, x2, x3, x4, x5


if __name__ == '__main__':
    pass



from unittest import TestCase

import torch

from base import VisionTestCase
from visage.matting.modnet import MODNetMatter


class ModNetTest(VisionTestCase):
    def test_modnet(self):

        # define hyper-parameters
        ref_size = 512

        image = self._get_obama_image(keep_aspect_ratio=True, resize=688, normalize=False)
        # image = image[:, int((image.shape[1] - 512) / 2):512 + int((image.shape[1] - 512) / 2)]

        modnet_matter = MODNetMatter()
        matte = modnet_matter.parse(image)
        print('hi')


#!/usr/bin/env python
import sys
from distutils.extension import Extension

import setuptools
from Cython.Distutils import build_ext  # Note, this requires Cython to be listed under "required" in pyproject.toml
from setuptools.command.build_ext import build_ext


# ----------------------------------------------------------
# Build procedure for compiled NMS module of FaceBoxesV2
# ----------------------------------------------------------

# run the customize_compiler
class custom_build_ext(build_ext):

    def finalize_options(self):
        build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        # __builtins__.__NUMPY_SETUP__ = False
        # Obtain the numpy include directory.  This logic works across numpy versions.
        import numpy as np

        try:
            numpy_include = np.get_include()
        except AttributeError:
            numpy_include = np.get_numpy_include()

        self.include_dirs.append(numpy_include)

    def build_extensions(self):
        # customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)


ext_modules = [
    Extension(
        "visage.modules.FaceBoxesV2.utils.nms.cpu_nms",
        ["src/visage/modules/FaceBoxesV2/utils/nms/cpu_nms.pyx"],
        extra_compile_args=["-Wno-cpp", "-Wno-unused-function"] if sys.platform == 'linux' else None,
    )
]

if __name__ == "__main__":
    setuptools.setup(
        include_package_data=True,
        # Important to also install .c/.cu/.hpp/.pyx files for nms
        # And .txt for meanfaces
        package_data={'': ['*.c', '*.pyx', '*.hpp', '*.cu', '*.txt']},

        # For building NMS module
        ext_modules=ext_modules,
        cmdclass={'build_ext': custom_build_ext},
        setup_requires=['numpy'],
    )



import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicConv2d(nn.Module):

    def __init__(self, in_channels, out_channels, **kwargs):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels, eps=1e-5)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return F.relu(x, inplace=True)


class Inception(nn.Module):

  def __init__(self):
    super(Inception, self).__init__()
    self.branch1x1 = BasicConv2d(128, 32, kernel_size=1, padding=0)
    self.branch1x1_2 = BasicConv2d(128, 32, kernel_size=1, padding=0)
    self.branch3x3_reduce = BasicConv2d(128, 24, kernel_size=1, padding=0)
    self.branch3x3 = BasicConv2d(24, 32, kernel_size=3, padding=1)
    self.branch3x3_reduce_2 = BasicConv2d(128, 24, kernel_size=1, padding=0)
    self.branch3x3_2 = BasicConv2d(24, 32, kernel_size=3, padding=1)
    self.branch3x3_3 = BasicConv2d(32, 32, kernel_size=3, padding=1)

  def forward(self, x):
    branch1x1 = self.branch1x1(x)

    branch1x1_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1)
    branch1x1_2 = self.branch1x1_2(branch1x1_pool)

    branch3x3_reduce = self.branch3x3_reduce(x)
    branch3x3 = self.branch3x3(branch3x3_reduce)

    branch3x3_reduce_2 = self.branch3x3_reduce_2(x)
    branch3x3_2 = self.branch3x3_2(branch3x3_reduce_2)
    branch3x3_3 = self.branch3x3_3(branch3x3_2)

    outputs = [branch1x1, branch1x1_2, branch3x3, branch3x3_3]
    return torch.cat(outputs, 1)


class CRelu(nn.Module):

  def __init__(self, in_channels, out_channels, **kwargs):
    super(CRelu, self).__init__()
    self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
    self.bn = nn.BatchNorm2d(out_channels, eps=1e-5)

  def forward(self, x):
    x = self.conv(x)
    x = self.bn(x)
    x = torch.cat([x, -x], 1)
    x = F.relu(x, inplace=True)
    return x


class FaceBoxes(nn.Module):

  def __init__(self, phase, size, num_classes):
    super(FaceBoxes, self).__init__()
    self.phase = phase
    self.num_classes = num_classes
    self.size = size

    self.conv1 = CRelu(3, 24, kernel_size=7, stride=4, padding=3)
    self.conv2 = CRelu(48, 64, kernel_size=5, stride=2, padding=2)

    self.inception1 = Inception()
    self.inception2 = Inception()
    self.inception3 = Inception()

    self.conv3_1 = BasicConv2d(128, 128, kernel_size=1, stride=1, padding=0)
    self.conv3_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)

    self.conv4_1 = BasicConv2d(256, 128, kernel_size=1, stride=1, padding=0)
    self.conv4_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)

    self.loc, self.conf = self.multibox(self.num_classes)

    if self.phase == 'test':
        self.softmax = nn.Softmax(dim=-1)

    if self.phase == 'train':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    nn.init.xavier_normal_(m.weight.data)
                    m.bias.data.fill_(0.02)
                else:
                    m.weight.data.normal_(0, 0.01)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

  def multibox(self, num_classes):
    loc_layers = []
    conf_layers = []
    loc_layers += [nn.Conv2d(128, 21 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(128, 21 * num_classes, kernel_size=3, padding=1)]
    loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
    loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
    return nn.Sequential(*loc_layers), nn.Sequential(*conf_layers)

  def forward(self, x):

    detection_sources = list()
    loc = list()
    conf = list()

    x = self.conv1(x)
    x = F.max_pool2d(x, kernel_size=3, stride=2, padding=1)
    x = self.conv2(x)
    x = F.max_pool2d(x, kernel_size=3, stride=2, padding=1)
    x = self.inception1(x)
    x = self.inception2(x)
    x = self.inception3(x)
    detection_sources.append(x)

    x = self.conv3_1(x)
    x = self.conv3_2(x)
    detection_sources.append(x)

    x = self.conv4_1(x)
    x = self.conv4_2(x)
    detection_sources.append(x)

    for (x, l, c) in zip(detection_sources, self.loc, self.conf):
        loc.append(l(x).permute(0, 2, 3, 1).contiguous())
        conf.append(c(x).permute(0, 2, 3, 1).contiguous())

    loc = torch.cat([o.view(o.size(0), -1) for o in loc], 1)
    conf = torch.cat([o.view(o.size(0), -1) for o in conf], 1)

    if self.phase == "test":
      output = (loc.view(loc.size(0), -1, 4),
                self.softmax(conf.view(conf.size(0), -1, self.num_classes)))
    else:
      output = (loc.view(loc.size(0), -1, 4),
                conf.view(conf.size(0), -1, self.num_classes))

    return output

class FaceBoxesV2Internal(nn.Module):

  def __init__(self, phase, size, num_classes):
    super(FaceBoxesV2Internal, self).__init__()
    self.phase = phase
    self.num_classes = num_classes
    self.size = size
    
    self.conv1 = BasicConv2d(3, 8, kernel_size=3, stride=2, padding=1)
    self.conv2 = BasicConv2d(8, 16, kernel_size=3, stride=2, padding=1)
    self.conv3 = BasicConv2d(16, 32, kernel_size=3, stride=2, padding=1)
    self.conv4 = BasicConv2d(32, 64, kernel_size=3, stride=2, padding=1)
    self.conv5 = BasicConv2d(64, 128, kernel_size=3, stride=2, padding=1)
    
    self.inception1 = Inception()
    self.inception2 = Inception()
    self.inception3 = Inception()
    
    self.conv6_1 = BasicConv2d(128, 128, kernel_size=1, stride=1, padding=0)
    self.conv6_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)
    
    self.conv7_1 = BasicConv2d(256, 128, kernel_size=1, stride=1, padding=0)
    self.conv7_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)
    
    self.loc, self.conf = self.multibox(self.num_classes)
    
    if self.phase == 'test':
        self.softmax = nn.Softmax(dim=-1)

    if self.phase == 'train':
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    nn.init.xavier_normal_(m.weight.data)
                    m.bias.data.fill_(0.02)
                else:
                    m.weight.data.normal_(0, 0.01)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

  def multibox(self, num_classes):
    loc_layers = []
    conf_layers = []
    loc_layers += [nn.Conv2d(128, 21 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(128, 21 * num_classes, kernel_size=3, padding=1)]
    loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
    loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
    conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
    return nn.Sequential(*loc_layers), nn.Sequential(*conf_layers)
    
  def forward(self, x):
  
    sources = list()
    loc = list()
    conf = list()

    x = self.conv1(x)
    x = self.conv2(x)
    x = self.conv3(x)
    x = self.conv4(x)
    x = self.conv5(x)
    x = self.inception1(x)
    x = self.inception2(x)
    x = self.inception3(x)
    sources.append(x)
    x = self.conv6_1(x)
    x = self.conv6_2(x)
    sources.append(x)
    x = self.conv7_1(x)
    x = self.conv7_2(x)
    sources.append(x)

    for (x, l, c) in zip(sources, self.loc, self.conf):
        loc.append(l(x).permute(0, 2, 3, 1).contiguous())
        conf.append(c(x).permute(0, 2, 3, 1).contiguous())
        
    loc = torch.cat([o.view(o.size(0), -1) for o in loc], 1)
    conf = torch.cat([o.view(o.size(0), -1) for o in conf], 1)

    if self.phase == "test":
      output = (loc.view(loc.size(0), -1, 4),
                self.softmax(conf.view(-1, self.num_classes)))
    else:
      output = (loc.view(loc.size(0), -1, 4),
                conf.view(conf.size(0), -1, self.num_classes))
  
    return output



import cv2
from torch import nn


class Detector(nn.Module):
    def __init__(self, model_arch, model_weights):
        super(Detector, self).__init__()
        self.model_arch = model_arch
        self.model_weights = model_weights

    def detect(self, image, thresh):
        raise NotImplementedError

    def crop(self, image, detections):
        crops = []
        for det in detections:
            xmin = max(det[2], 0)
            ymin = max(det[3], 0)
            width = det[4]
            height = det[5]
            xmax = min(xmin+width, image.shape[1])
            ymax = min(ymin+height, image.shape[0])
            cut = image[ymin:ymax, xmin:xmax,:]
            crops.append(cut)

        return crops

    def draw(self, image, detections, im_scale=None):
        if im_scale is not None:
            image = cv2.resize(image, None, None, fx=im_scale, fy=im_scale, interpolation=cv2.INTER_LINEAR)
            detections = [[det[0],det[1],int(det[2]*im_scale),int(det[3]*im_scale),int(det[4]*im_scale),int(det[5]*im_scale)] for det in detections]

        for det in detections:
            xmin = det[2]
            ymin = det[3]
            width = det[4]
            height = det[5]
            xmax = xmin + width
            ymax = ymin + height
            cv2.rectangle(image, (xmin, ymin), (xmax, ymax), (0, 0, 255), 2)

        return image



import logging
import os

import numpy as np
import torch
from scipy.integrate import simps


def get_label(data_name, label_file, task_type=None):
    label_path = os.path.join('data', data_name, label_file)
    with open(label_path, 'r') as f:
        labels = f.readlines()
    labels = [x.strip().split() for x in labels]
    if len(labels[0]) == 1:
        return labels

    labels_new = []
    for label in labels:
        image_name = label[0]
        target = label[1:]
        target = np.array([float(x) for x in target])
        if task_type is None:
            labels_new.append([image_name, target])
        else:
            labels_new.append([image_name, task_type, target])
    return labels_new


def get_meanface(meanface_file, num_nb):
    with open(meanface_file) as f:
        meanface = f.readlines()[0]

    meanface = meanface.strip().split()
    meanface = [float(x) for x in meanface]
    meanface = np.array(meanface).reshape(-1, 2)
    # each landmark predicts num_nb neighbors
    meanface_indices = []
    for i in range(meanface.shape[0]):
        pt = meanface[i, :]
        dists = np.sum(np.power(pt - meanface, 2), axis=1)
        indices = np.argsort(dists)
        meanface_indices.append(indices[1:1 + num_nb])

    # each landmark predicted by X neighbors, X varies
    meanface_indices_reversed = {}
    for i in range(meanface.shape[0]):
        meanface_indices_reversed[i] = [[], []]
    for i in range(meanface.shape[0]):
        for j in range(num_nb):
            meanface_indices_reversed[meanface_indices[i][j]][0].append(i)
            meanface_indices_reversed[meanface_indices[i][j]][1].append(j)

    max_len = 0
    for i in range(meanface.shape[0]):
        tmp_len = len(meanface_indices_reversed[i][0])
        if tmp_len > max_len:
            max_len = tmp_len

    # tricks, make them have equal length for efficient computation
    for i in range(meanface.shape[0]):
        tmp_len = len(meanface_indices_reversed[i][0])
        meanface_indices_reversed[i][0] += meanface_indices_reversed[i][0] * 10
        meanface_indices_reversed[i][1] += meanface_indices_reversed[i][1] * 10
        meanface_indices_reversed[i][0] = meanface_indices_reversed[i][0][:max_len]
        meanface_indices_reversed[i][1] = meanface_indices_reversed[i][1][:max_len]

    # make the indices 1-dim
    reverse_index1 = []
    reverse_index2 = []
    for i in range(meanface.shape[0]):
        reverse_index1 += meanface_indices_reversed[i][0]
        reverse_index2 += meanface_indices_reversed[i][1]
    return meanface_indices, reverse_index1, reverse_index2, max_len


def compute_loss_pip(outputs_map, outputs_local_x, outputs_local_y, outputs_nb_x, outputs_nb_y, labels_map,
                     labels_local_x, labels_local_y, labels_nb_x, labels_nb_y, criterion_cls, criterion_reg, num_nb):
    tmp_batch, tmp_channel, tmp_height, tmp_width = outputs_map.size()
    labels_map = labels_map.view(tmp_batch * tmp_channel, -1)
    labels_max_ids = torch.argmax(labels_map, 1)
    labels_max_ids = labels_max_ids.view(-1, 1)
    labels_max_ids_nb = labels_max_ids.repeat(1, num_nb).view(-1, 1)

    outputs_local_x = outputs_local_x.view(tmp_batch * tmp_channel, -1)
    outputs_local_x_select = torch.gather(outputs_local_x, 1, labels_max_ids)
    outputs_local_y = outputs_local_y.view(tmp_batch * tmp_channel, -1)
    outputs_local_y_select = torch.gather(outputs_local_y, 1, labels_max_ids)
    outputs_nb_x = outputs_nb_x.view(tmp_batch * num_nb * tmp_channel, -1)
    outputs_nb_x_select = torch.gather(outputs_nb_x, 1, labels_max_ids_nb)
    outputs_nb_y = outputs_nb_y.view(tmp_batch * num_nb * tmp_channel, -1)
    outputs_nb_y_select = torch.gather(outputs_nb_y, 1, labels_max_ids_nb)

    labels_local_x = labels_local_x.view(tmp_batch * tmp_channel, -1)
    labels_local_x_select = torch.gather(labels_local_x, 1, labels_max_ids)
    labels_local_y = labels_local_y.view(tmp_batch * tmp_channel, -1)
    labels_local_y_select = torch.gather(labels_local_y, 1, labels_max_ids)
    labels_nb_x = labels_nb_x.view(tmp_batch * num_nb * tmp_channel, -1)
    labels_nb_x_select = torch.gather(labels_nb_x, 1, labels_max_ids_nb)
    labels_nb_y = labels_nb_y.view(tmp_batch * num_nb * tmp_channel, -1)
    labels_nb_y_select = torch.gather(labels_nb_y, 1, labels_max_ids_nb)

    labels_map = labels_map.view(tmp_batch, tmp_channel, tmp_height, tmp_width)
    loss_map = criterion_cls(outputs_map, labels_map)
    loss_x = criterion_reg(outputs_local_x_select, labels_local_x_select)
    loss_y = criterion_reg(outputs_local_y_select, labels_local_y_select)
    loss_nb_x = criterion_reg(outputs_nb_x_select, labels_nb_x_select)
    loss_nb_y = criterion_reg(outputs_nb_y_select, labels_nb_y_select)
    return loss_map, loss_x, loss_y, loss_nb_x, loss_nb_y


def train_model(det_head, net, train_loader, criterion_cls, criterion_reg, cls_loss_weight, reg_loss_weight, num_nb,
                optimizer, num_epochs, scheduler, save_dir, save_interval, device):
    for epoch in range(num_epochs):
        print('Epoch {}/{}'.format(epoch, num_epochs - 1))
        logging.info('Epoch {}/{}'.format(epoch, num_epochs - 1))
        print('-' * 10)
        logging.info('-' * 10)
        net.train()
        epoch_loss = 0.0

        for i, data in enumerate(train_loader):
            if det_head == 'pip':
                inputs, labels_map, labels_x, labels_y, labels_nb_x, labels_nb_y = data
                inputs = inputs.to(device)
                labels_map = labels_map.to(device)
                labels_x = labels_x.to(device)
                labels_y = labels_y.to(device)
                labels_nb_x = labels_nb_x.to(device)
                labels_nb_y = labels_nb_y.to(device)
                outputs_map, outputs_x, outputs_y, outputs_nb_x, outputs_nb_y = net(inputs)
                loss_map, loss_x, loss_y, loss_nb_x, loss_nb_y = compute_loss_pip(outputs_map, outputs_x, outputs_y,
                                                                                  outputs_nb_x, outputs_nb_y,
                                                                                  labels_map, labels_x, labels_y,
                                                                                  labels_nb_x, labels_nb_y,
                                                                                  criterion_cls, criterion_reg, num_nb)
                loss = cls_loss_weight * loss_map + reg_loss_weight * loss_x + reg_loss_weight * loss_y + reg_loss_weight * loss_nb_x + reg_loss_weight * loss_nb_y
            else:
                print('No such head:', det_head)
                exit(0)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if i % 10 == 0:
                if det_head == 'pip':
                    print(
                        '[Epoch {:d}/{:d}, Batch {:d}/{:d}] <Total loss: {:.6f}> <map loss: {:.6f}> <x loss: {:.6f}> <y loss: {:.6f}> <nbx loss: {:.6f}> <nby loss: {:.6f}>'.format(
                            epoch, num_epochs - 1, i, len(train_loader) - 1, loss.item(),
                                   cls_loss_weight * loss_map.item(), reg_loss_weight * loss_x.item(),
                                   reg_loss_weight * loss_y.item(), reg_loss_weight * loss_nb_x.item(),
                                   reg_loss_weight * loss_nb_y.item()))
                    logging.info(
                        '[Epoch {:d}/{:d}, Batch {:d}/{:d}] <Total loss: {:.6f}> <map loss: {:.6f}> <x loss: {:.6f}> <y loss: {:.6f}> <nbx loss: {:.6f}> <nby loss: {:.6f}>'.format(
                            epoch, num_epochs - 1, i, len(train_loader) - 1, loss.item(),
                                   cls_loss_weight * loss_map.item(), reg_loss_weight * loss_x.item(),
                                   reg_loss_weight * loss_y.item(), reg_loss_weight * loss_nb_x.item(),
                                   reg_loss_weight * loss_nb_y.item()))
                else:
                    print('No such head:', det_head)
                    exit(0)
            epoch_loss += loss.item()
        epoch_loss /= len(train_loader)
        if epoch % (save_interval - 1) == 0 and epoch > 0:
            filename = os.path.join(save_dir, 'epoch%d.pth' % epoch)
            torch.save(net.state_dict(), filename)
            print(filename, 'saved')
        scheduler.step()
    return net


def forward_pip(net, inputs, preprocess, input_size, net_stride, num_nb):
    net.eval()

    with torch.no_grad():
        outputs_cls, outputs_x, outputs_y, outputs_nb_x, outputs_nb_y = net(inputs)
        if not hasattr(net, 'cnt'):
            net.cnt = 0
        # rnd_idx = np.random.randint(0, 97)
        # plt.imshow(outputs_cls.detach().cpu().numpy()[0, rnd_idx, :, :])
        # plt.show()
        # np.save('/mnt/hdd/heatmaps_{:03d}.npy'.format(net.cnt), outputs_cls.detach().cpu().numpy())
        net.cnt += 1

        tmp_batch, tmp_channel, tmp_height, tmp_width = outputs_cls.size()
        assert tmp_batch == 1

        outputs_cls = outputs_cls.view(tmp_batch * tmp_channel, -1)
        max_ids = torch.argmax(outputs_cls, 1)
        max_cls = torch.max(outputs_cls, 1)[0]
        max_ids = max_ids.view(-1, 1)
        max_ids_nb = max_ids.repeat(1, num_nb).view(-1, 1)

        outputs_x = outputs_x.view(tmp_batch * tmp_channel, -1)
        outputs_x_select = torch.gather(outputs_x, 1, max_ids)
        outputs_x_select = outputs_x_select.squeeze(1)
        outputs_y = outputs_y.view(tmp_batch * tmp_channel, -1)
        outputs_y_select = torch.gather(outputs_y, 1, max_ids)
        outputs_y_select = outputs_y_select.squeeze(1)

        outputs_nb_x = outputs_nb_x.view(tmp_batch * num_nb * tmp_channel, -1)
        outputs_nb_x_select = torch.gather(outputs_nb_x, 1, max_ids_nb)
        outputs_nb_x_select = outputs_nb_x_select.squeeze(1).view(-1, num_nb)
        outputs_nb_y = outputs_nb_y.view(tmp_batch * num_nb * tmp_channel, -1)
        outputs_nb_y_select = torch.gather(outputs_nb_y, 1, max_ids_nb)
        outputs_nb_y_select = outputs_nb_y_select.squeeze(1).view(-1, num_nb)

        tmp_x = (max_ids % tmp_width).view(-1, 1).float() + outputs_x_select.view(-1, 1)
        tmp_y = (max_ids // tmp_width).view(-1, 1).float() + outputs_y_select.view(-1, 1)
        tmp_x /= 1.0 * input_size / net_stride
        tmp_y /= 1.0 * input_size / net_stride

        tmp_nb_x = (max_ids % tmp_width).view(-1, 1).float() + outputs_nb_x_select
        tmp_nb_y = (max_ids // tmp_width).view(-1, 1).float() + outputs_nb_y_select
        tmp_nb_x = tmp_nb_x.view(-1, num_nb)
        tmp_nb_y = tmp_nb_y.view(-1, num_nb)
        tmp_nb_x /= 1.0 * input_size / net_stride
        tmp_nb_y /= 1.0 * input_size / net_stride

    return tmp_x, tmp_y, tmp_nb_x, tmp_nb_y, outputs_cls, max_cls


def compute_nme(lms_pred, lms_gt, norm):
    lms_pred = lms_pred.reshape((-1, 2))
    lms_gt = lms_gt.reshape((-1, 2))
    nme = np.mean(np.linalg.norm(lms_pred - lms_gt, axis=1)) / norm
    return nme


def compute_fr_and_auc(nmes, thres=0.1, step=0.0001):
    num_data = len(nmes)
    xs = np.arange(0, thres + step, step)
    ys = np.array([np.count_nonzero(nmes <= x) for x in xs]) / float(num_data)
    fr = 1.0 - ys[-1]
    auc = simps(ys, x=xs) / thres
    return fr, auc
